from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


__all__ = ['Matrix']


class MatrixError(Exception):
    pass


class Matrix(object):
    """"""

    dtypes = {
        'i': int,
        'int': int,
        'f': float,
        'float': float,
    }

    __slots__ = ['_data', '_dtype']

    def __init__(self, data, dtype=None):
        self._data = None
        self._dtype = None
        self.dtype = dtype
        self.data = data

    @property
    def data(self):
        """"""
        return self._data

    @data.setter
    def data(self, data):
        self._data = []

        dtype = self.dtypes[self.dtype]

        if isinstance(data, list):
            cols = None

            for row in data:
                if isinstance(row, (list, tuple)):
                    if not cols:
                        cols = len(row)
                    if len(row) != cols:
                        raise MatrixError('The shape of the data is not rectangular.')
                    self._data.append([dtype(value) for value in row])
                else:
                    raise MatrixError('The data is not two-dimensional.')

    @property
    def dtype(self):
        """"""
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        if not dtype:
            dtype = 'float'
        dtype = str(dtype)
        if dtype not in Matrix.dtypes:
            raise MatrixError('Data type not supported.')
        self._dtype = dtype

    @property
    def shape(self):
        """"""
        return len(self._data), len(self._data[0])

    @property
    def size(self):
        """"""
        m, n = self.shape
        return m * n

    def __str__(self):
        """"""
        return "Matrix({})".format(self._data)

    def __getitem__(self, key):
        """"""
        if isinstance(key, slice):
            return Matrix([self._data[i] for i in range(* key.indices(self.shape[0]))])

        if isinstance(key, list):
            return Matrix([self._data[i] for i in key])

        if isinstance(key, int):
            return Matrix([self._data[key]])

        if isinstance(key, tuple):
            i, j = key
            if isinstance(i, int) and isinstance(j, int):
                return self._data[i][j]
            return self[key[0]].transpose()[key[1]].transpose()

        raise KeyError

    def transpose(self):
        """"""
        return Matrix(zip(*self._data))

    def to_list(self):
        """"""
        pass

    def to_array(self):
        """"""
        pass

    def reshape(self, shape, order):
        """"""
        pass


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    m = Matrix([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    print(m[0])
    print(m[0, 1])
    print(m[[0, 1]])
    print(m[0:2])
    print(m[0, 0:2])
    print(m[0:2, 0])
    print(m[0:2, 0:2])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from compas_tna.diagrams import FormDiagram
from compas.numerical import fd_numpy


__all__ = [
    'relax_boundary_openings',
    'relax_boundary_openings_proxy',
]


def relax_boundary_openings_proxy(formdata, fixed):
    form = FormDiagram.from_data(formdata)
    relax_boundary_openings(form, fixed)
    return form.to_data()


def relax_boundary_openings(form, fixed):
    k_i = form.key_index()
    xyz = form.vertices_attributes('xyz')
    edges = [(k_i[u], k_i[v]) for u, v in form.edges()]
    fixed = [k_i[key] for key in fixed]
    q = form.edges_attribute('q')
    loads = form.vertices_attributes(('px', 'py', 'pz'))
    xyz, q, f, l, r = fd_numpy(xyz, edges, fixed, q, loads)
    for key in form.vertices():
        index = k_i[key]
        form.vertex_attributes(key, 'xyz', xyz[index])


# ============================================================================
# Main
# ============================================================================

if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ._array import *
from ._sparsearray import *
from ._matrix import *


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_tna.diagrams import Diagram


__all__ = ['ForceDiagram']


class ForceDiagram(Diagram):
    """The ``ForceDiagram`` defines a TNA force diagram.

    Attributes
    ----------
    primal : :class:`compas_tna.diagrams.FormDiagram`
        The TNA form diagram corresponding to the force diagram.
    scale : float
        The scale factor for calculating horizontal forces based on edge lengths.
        The horizontal force in an edge of the form diagram is equal
        to the length of the corresponding edge in the force diagram multiplied by the scale factor.
    default_vertex_attributes : dict
        The names and default values of the attributes of vertices.
    default_edge_attributes : dict
        The names and default values of the attributes of the edges.
    default_face_attributes : dict
        The names and default values of the attributes of the faces.

    Notes
    -----
    In TNA, a pair of form and force diagrams describe the horizontal equilibrium of a 3D network of forces
    with vertical loads applied to its nodes, if they are reciprocal.
    This means they are each others dual and that corresponding edge pairs are at a constant angle.
    Typically this angle is 0 or 90 degrees, but any other constant angle is sufficient.

    When the force diagram is created from the form diagram, both diagrams are dual.
    However, the reciprocal relationship has to be established separately,
    using one of the horizontal equilibrium solvers.

    """

    def __init__(self):
        super(ForceDiagram, self).__init__()
        self.primal = None
        self.scale = 1.0
        self.default_vertex_attributes.update({
            'x': 0.0,
            'y': 0.0,
            'z': 0.0,
            'is_fixed': False,
        })
        self.default_edge_attributes.update({
            'lmin': 0.0,
            'lmax': 1e+7,

            '_a': 0.0,
        })
        self.attributes.update({
            'name': 'ForceDiagram',
            'scale': 1.0,

            'color.vertex': (255, 255, 255),
            'color.edge': (0, 0, 0),
            'color.face': (210, 210, 210),
        })

    # --------------------------------------------------------------------------
    # Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_formdiagram(cls, formdiagram):
        """Construct a force diagram from a given form diagram.

        Parameters
        ----------
        formdiagram : :class:`compas_tna.diagrams.FormDiagram`
            A form diagram instance.

        Returns
        -------
        :class:`compas_tna.diagrams.ForceDiagram`
            The dual force diagram.

        """
        dual = formdiagram.dual_diagram(cls)
        dual.vertices_attribute('z', 0.0)
        dual.primal = formdiagram
        formdiagram.dual = dual
        return dual

    # --------------------------------------------------------------------------
    # Vertices
    # --------------------------------------------------------------------------

    def fixed(self):
        """Vertices with ``is_fixed`` set to ``True``.

        Returns
        -------
        generator
            A generator object for iteration over vertex keys that are fixed.
        """
        return self.vertices_where({'is_fixed': True})

    # --------------------------------------------------------------------------
    # Helpers
    # --------------------------------------------------------------------------

    def uv_index(self, form=None):
        """Construct a map relating edge uv pairs to their index in an edge list.

        Parameters
        ----------
        form : :class:`compas_tna.diagrams.FormDiagram`, optional
            If provided, this maps edge uv's to the index in a list matching the ordering of corresponding edges in the form diagram.

        Returns
        -------
        dict
            A dict mapping edge uv tuples to indices in an (ordered) edge list.
        """
        if not form:
            return {uv: index for index, uv in enumerate(self.edges())}
        uv_index = dict()
        for index, (u, v) in enumerate(form.edges_where({'_is_edge': True})):
            f1 = form.halfedge[u][v]
            f2 = form.halfedge[v][u]
            uv_index[(f1, f2)] = index
        return uv_index

    def ordered_edges(self, form):
        """Construct an edge list in which the edges are ordered according to the ordering of edges in a corresponding list of form diagram edges.

        Parameters
        ----------
        form : :class:`compas_tna.diagrams.FormDiagram`, optional
            The form diagram according to which the edges should be ordered.

        Returns
        -------
        list
            A list of edge uv tuples.
        """
        uv_index = self.uv_index(form=form)
        index_uv = {index: uv for uv, index in iter(uv_index.items())}
        edges = [index_uv[index] for index in range(self.number_of_edges())]
        return edges

    def get_form_edge_attribute(self, form, key, name, value=None):
        f1, f2 = key
        for u, v in form.face_halfedges(f1):
            if form.halfedge[v][u] == f2:
                break
        return form.edge_attribute((u, v), name, value=value)

    # --------------------------------------------------------------------------
    # visualisation
    # --------------------------------------------------------------------------

    def plot(self):
        """Plot a force diagram with a plotter with all the default settings."""
        from compas_plotters import MeshPlotter
        plotter = MeshPlotter(self, figsize=(12, 8), tight=True)
        plotter.draw_vertices(radius=0.05)
        plotter.draw_edges()
        plotter.show()

    def draw(self, layer=None, clear_layer=True, settings=None):
        """Draw the force diagram in Rhino.

        Parameters
        ----------
        layer : str, optional
            The layer in which the drawing should be contained.
        clear_layer : bool, optional
            Clear the layer if ``True``.
            Default is ``True``.
        settings : dict, optional
            A dictionary of settings overwriting the default settings of the artist.
        """
        from compas_tna.rhino import ForceArtist
        artist = ForceArtist(self, layer=layer)
        if clear_layer:
            artist.clear_layer()
        if not settings:
            settings = {}
        vertexcolor = {}
        vertexcolor.update({key: '#00ff00' for key in self.vertices_where({'is_fixed': True})})
        artist.draw_vertices(color=vertexcolor)
        artist.draw_edges()
        artist.redraw()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import array
import compas

from compas.utilities import flatten

try:
    from compas.numerical._alglib._core import xalglib
except ImportError:
    compas.raise_if_ironpython()


__all__ = ['Array', 'Zeros', 'Ones', 'Diagonal', 'Eye', 'ZerosLike']


class ArrayError(Exception):
    pass


class Array(object):
    """"""

    dtypes = {
        'i': 'i',
        'int': 'i',
        'f': 'f',
        'float': 'f',
    }

    __slots__ = ['_data', '_dtype', '_shape']

    def __init__(self, data, shape, dtype=None):
        self._data = None
        self._dtype = None
        self._shape = None
        self.dtype = dtype
        self.data = data
        self.shape = shape

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        if not dtype:
            dtype = 'float'
        dtype = str(dtype)
        if dtype not in Array.dtypes:
            raise ArrayError('Data type not supported.')
        self._dtype = Array.dtypes[dtype]

    @property
    def data(self):
        m, n = self.shape
        data = self._data.tolist()
        if not n:
            return data
        return [data[i * n: (i + 1) * n] for i in range(m)]

    @data.setter
    def data(self, data):
        if isinstance(data, array.array):
            self._data = data
        else:
            self._data = array.array(self.dtype)

            if isinstance(data[0], (array.array, list, tuple)):
                data = list(flatten(data))

            self._data.fromlist(list(data))

    @property
    def flatdata(self):
        return self._data.tolist()

    @property
    def size(self):
        """"""
        return len(self._data)

    @property
    def shape(self):
        """"""
        return self._shape

    @shape.setter
    def shape(self, shape):
        m, n = shape
        if not n:
            n = 1
        if m * n != self.size:
            raise ArrayError('Size and shape are not compatible.')
        self._shape = shape

    @property
    def rows(self):
        m, n = self.shape
        data = self._data.tolist()
        if not n:
            return data
        return [data[i * n: (i + 1) * n] for i in range(m)]

    @property
    def cols(self):
        m, n = self.shape
        data = self._data.tolist()
        if not n:
            return data
        return [[data[i + j * n] for j in range(m)] for i in range(n)]

    # ==========================================================================
    # customisation
    # ==========================================================================

    def __len__(self):
        return self.shape[0]

    def __str__(self):
        return "Array({})".format(self.data)

    def __iter__(self):
        return iter(self.rows)

    # ==========================================================================
    # access
    # ==========================================================================

    def __getitem__(self, key):
        m, n = self.shape

        if isinstance(key, int):
            if key > m - 1:
                raise KeyError
            if not n:
                return self._data[m]
            return Array(self._data[key * n: (key + 1) * n], (n, None), self.dtype)

        if isinstance(key, list):
            if not n:
                data = [self._data[i] for i in key]
            else:
                data = [self._data[i * n: (i + 1) * n] for i in key]
            return Array(data, (len(key), n), self.dtype)

        if isinstance(key, slice):
            keys = list(range(* key.indices(m)))
            return self[keys]

        if isinstance(key, tuple):
            i, j = key
            i_int = isinstance(i, int)
            j_int = isinstance(j, int)
            if i_int and j_int:
                return self._data[i * n + j]
            if i_int:
                return self[i][j]
            if j_int:
                return self[i].transpose()[j]
            return self[i].transpose()[j].transpose()

        raise KeyError

    # ==========================================================================
    # modification
    # ==========================================================================

    def __setitem__(self, key, value):
        m, n = self.shape

        if not n:
            n = 1

        if isinstance(key, int):
            if n == 1:
                self._data[key] = value
            else:
                if key >= m:
                    raise KeyError
                if isinstance(value, Array):
                    a, b = value.shape
                    if b and b > 1:
                        raise ValueError
                    if a != n:
                        raise ValueError
                    for i, v in enumerate(value._data):
                        self._data[key * n + i] = v
                else:
                    if len(value) != n:
                        raise ValueError
                    for i, v in enumerate(value):
                        self._data[key * n + i] = v

        elif isinstance(key, list):
            for i, k in enumerate(key):
                self[k] = value[i]

        elif isinstance(key, slice):
            keys = list(range(* key.indices(m)))
            self[keys] = value

        elif isinstance(key, tuple):
            i, j = key
            i_int = isinstance(i, int)
            j_int = isinstance(j, int)
            if i_int and j_int:
                if i >= m or j >= n:
                    raise KeyError
                self._data[i * n + j] = value

            elif i_int:
                if isinstance(j, slice):
                    j = list(range(* j.indices(m)))
                if len(j) != len(value):
                    raise ValueError(
                        "Sizes of row slice and replacement values don't match: {} != {}".format(len(j), len(value)))
                for index, data in zip(j, value):
                    self._data[i * n + index] = data

            elif j_int:
                if isinstance(i, slice):
                    i = list(range(* i.indices(m)))
                if len(i) != len(value):
                    raise ValueError(
                        "Sizes of column slice and replacement values don't match: {} != {}".format(len(i), len(value)))
                for index, data in zip(i, value):
                    self._data[index * n + j] = data

            else:
                raise NotImplementedError

        else:
            raise KeyError

    # ==========================================================================
    # helpers
    # ==========================================================================

    # rename to "flatdata"
    def flatten(self):
        return self._data.tolist()

    def transpose(self):
        m, n = self.shape
        if not n:
            n = 1
        return Array([self._data[i + j * n] for i in range(n) for j in range(m)], (n, m), self.dtype)

    # ==========================================================================
    # linalg
    # ==========================================================================

    def add(self, other):
        m, n = self.shape
        o, p = other.shape
        if m != o or n != p:
            raise ArrayError('Incompatible matrix dimensions.')
        data = [a + b for a, b in zip(self._data, other._data)]
        return Array(data, (m, n), dtype=self.dtype)

    def subtract(self, other):
        m, n = self.shape
        o, p = other.shape
        if m != o or n != p:
            raise ArrayError('Incompatible matrix dimensions.')
        data = [a - b for a, b in zip(self._data, other._data)]
        return Array(data, (m, n), dtype=self.dtype)

    def multiply(self, factor):
        pass

    def emultiply(self, other):
        pass

    def _dot(self, other):
        a, b = self.shape
        c, d = other.shape
        if b != c:
            raise ArrayError('Incompatible shapes for dot product.')

        rows = self.rows
        cols = other.cols
        m, n = a, d
        data = [None] * (m * n)
        for i in range(m):
            for j in range(n):
                data[i * n + j] = sum(x * y for x, y in zip(rows[i], cols[j]))
        return Array(data, (m, n), self.dtype)

    def dot(self, other):
        try:
            xalglib
        except NameError:
            return self._dot(other)

        m, k = self.shape
        k, n = other.shape

        data = xalglib.rmatrixgemm(
            m, n, k,
            1.0, self.data, 0, 0, 0,
            other.data, 0, 0, 0,
            0, Zeros((m, n)).data, 0, 0
        )
        return Array(data, (m, n), self.dtype)

    def tdot(self, other):
        try:
            xalglib
        except NameError:
            return self.transpose()._dot(other)

        k, m = self.shape
        k, n = other.shape

        data = xalglib.rmatrixgemm(
            m, n, k,
            1.0, self.data, 0, 0, 1,
            other.data, 0, 0, 0,
            0, Zeros((m, n)).data, 0, 0
        )
        return Array(data, (m, n), self.dtype)

    def xdot(self, B, C, a=1.0, c=1.0):
        A = self
        try:
            xalglib
        except NameError:
            return A.multiply(a)._dot(B).add(C.multiply(c))

        m, k = A.shape
        k, n = B.shape

        data = xalglib.rmatrixgemm(
            m, n, k,
            a, A.data, 0, 0, 0,
            B.data, 0, 0, 0,
            c, C.data, 0, 0
        )

        return Array(data, (m, n), self.dtype)


# ==============================================================================
# Special
# ==============================================================================


class Zeros(Array):

    def __init__(self, shape, dtype='f'):
        m, n = shape
        if not n:
            n = 1
        data = [0.0] * (m * n)
        super(Zeros, self).__init__(data, shape, dtype)


class Ones(Array):

    def __init__(self, shape, dtype='f'):
        m, n = shape
        if not n:
            n = 1
        data = [1.0] * (m * n)
        super(Ones, self).__init__(data, shape, dtype)


class Diagonal(Zeros):

    def __init__(self, data, dtype='f'):
        m = len(data)
        super(Diagonal, self).__init__((m, m), dtype)
        for i in range(m):
            self._data[i * (m + 1)] = data[i]


class Eye(Diagonal):

    def __init__(self, shape, dtype='f'):
        data = [1.0] * shape
        super(Eye, self).__init__(data, dtype)


class ZerosLike(Array):

    def __init__(self, array):
        m, n = array.shape
        data = [0.0] * (m * n)
        super(ZerosLike, self).__init__(data, array.shape, array.dtype)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    data = range(9)

    a = Array(data, (3, 3), dtype='i')

    for row in a:
        print(row)

    print(a)
    print(a[0, :])
    print(a[:, 0])

    print(a[[0, 1]])
    print(a[:, [0, 1]])

    a[[1, 2], 0] = [1, 1]
    print(a)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

try:
    from compas.numerical._alglib._core import xalglib
    from compas.numerical._alglib._core._array import Array
    from compas.numerical._alglib._core._array import Zeros
except ImportError:
    compas.raise_if_ironpython()


__all__ = ['SparseArray', 'SparseDiagonal']


class SparseArrayError(Exception):
    pass


class SparseArray(object):

    dtypes = {
        'f': float,
        'float': float,
        'i': int,
        'int': int
    }

    def __init__(self, ijk, shape, dtype='f'):
        self.__rows = None
        self.__cols = None
        self.__data = None
        self._data = None
        self._shape = None
        self._dtype = None
        self.shape = shape
        self.dtype = dtype
        self.data = ijk

    # ==========================================================================
    # descriptors
    # ==========================================================================

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, ijk):
        m, n = self.shape
        dtype = SparseArray.dtypes[self.dtype]
        rows, cols, data = ijk
        self.__rows = rows
        self.__cols = cols
        self.__data = data
        self._data = {i: {} for i in rows}
        for i, j, k in zip(rows, cols, data):
            if i >= m or j >= n:
                raise SparseArrayError('Data not compatible with shape.')
            self._data[i][j] = dtype(k)

    @property
    def matrix(self):
        m, n = self.shape
        M = xalglib.sparsecreate(m, n, len(self.__data))
        for i in self._data:
            for j in self._data[i]:
                k = self._data[i][j]
                xalglib.sparseset(M, i, j, k)
        xalglib.sparseconverttocrs(M)
        return M

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        if dtype not in SparseArray.dtypes:
            raise TypeError
        self._dtype = dtype

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        m, n = shape
        self._shape = m, n

    # ==========================================================================
    # customisation
    # ==========================================================================

    def __str__(self):
        return str(self._data)

    def __getitem__(self, key):
        m, n = self.shape

        if isinstance(key, int):
            row = []
            if key >= m:
                raise KeyError
            for j in range(n):
                if key in self._data and j in self._data[key]:
                    row.append(self._data[key][j])
                else:
                    row.append(0.0)
            return row

        if isinstance(key, list):
            keys = key
            rows, cols, data = [], [], []
            for i, key in enumerate(keys):
                if key >= m:
                    raise KeyError
                if key in self._data:
                    for j in self._data[key]:
                        rows.append(i)
                        cols.append(j)
                        data.append(self._data[key][j])
            return SparseArray((rows, cols, data), (len(keys), n))

        if isinstance(key, slice):
            keys = range(* key.indices(m))
            return self[keys]

        if isinstance(key, tuple):
            i, j = key
            i_int = isinstance(i, int)
            j_int = isinstance(j, int)
            if i_int and j_int:
                if i in self._data and j in self._data[i]:
                    return self._data[i][j]
                return 0.0
            if i_int:
                return self[i][j]
            if j_int:
                return self[i].transpose()[j]
            return self[i].transpose()[j].transpose()

    def __setitem__(self, key, value):
        raise NotImplementedError

    # ==========================================================================
    # conversions
    # ==========================================================================

    def to_dense(self):
        m, n = self.shape
        data = []
        for i in range(m):
            for j in range(n):
                if i in self._data and j in self._data[i]:
                    data.append(self._data[i][j])
                else:
                    data.append(0.0)
        return Array(data, (m, n))

    def to_csc(self):
        pass

    # ==========================================================================
    # linalg
    # ==========================================================================

    def diagonal(self):
        m, n = self.shape
        if m <= n:
            return [self[i, i] for i in range(m)]
        return [self[i, i] for i in range(n)]

    def transpose(self):
        m, n = self.shape
        rows, cols, data = [], [], []
        for i in self._data:
            for j in self._data[i]:
                k = self._data[i][j]
                rows.append(j)
                cols.append(i)
                data.append(k)
        return SparseArray((rows, cols, data), (n, m))

    def _dot(self, other):
        A = self
        B = other.transpose()
        m, n = A.shape
        k, n = B.shape
        shape = m, k
        rows, cols, data = [], [], []
        Arows = {i: set(A._data[i]) for i in A._data}
        Bcols = {j: set(B._data[j]) for j in B._data}
        for i, setrow in iter(Arows.items()):
            for j, setcol in iter(Bcols.items()):
                keys = setrow & setcol
                if keys:
                    value = sum(A._data[i][key] * B._data[j][key] for key in keys)
                    rows.append(i)
                    cols.append(j)
                    data.append(value)
        return SparseArray((rows, cols, data), shape)

    def _tdot(self, other):
        A = self.transpose()
        B = other
        return A._dot(B)

    def dot(self, other):
        if isinstance(other, Array):
            m, n = self.shape
            n, k = other.shape
            A = self.matrix
            b = other.data
            data = xalglib.sparsemm(A, b, k, Zeros((m, k)).data)
            return Array(data, (m, k))
        return self._dot(other)

    def tdot(self, other):
        if isinstance(other, Array):
            m, n = self.shape
            n, k = other.shape
            A = self.matrix
            b = other.data
            data = xalglib.sparsemtm(A, b, k, Zeros((n, k)).data)
            return Array(data, (n, k))
        return self._tdot(other)

    def xdot(self, B, C, a=1.0, c=1.0):
        raise NotImplementedError


# ==============================================================================
# Special
# ==============================================================================


class SparseDiagonal(SparseArray):

    def __init__(self, data, dtype='f'):
        m = len(data)
        rows = range(m)
        super(SparseDiagonal, self).__init__((rows, rows, data), (m, m), dtype)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    import compas
    from compas.datastructures import Network

    network = Network.from_obj(compas.get('lines.obj'))

    e = network.number_of_edges()
    v = network.number_of_vertices()

    vertices = network.vertices_attributes('xyz')
    edges = list(network.edges())

    xyz = Array(vertices, (v, 3))

    shape = e, v

    rows, cols, data = [], [], []

    for i, (u, v) in enumerate(edges):
        rows.append(i)
        rows.append(i)
        cols.append(u)
        cols.append(v)
        data.append(-1)
        data.append(+1)

    C = SparseArray((rows, cols, data), shape)

    print(C.transpose()._dot(C).diagonal())
    print(C.transpose().dot(C).diagonal())
    print(C.tdot(C).diagonal())

    # print(C[:, 0])
    # print(Ct[0, :])

    # print(C[:, [0, 1, 2]])
    # print(Ct[[0, 1, 2], :])

    # print(C[0])
    # print(C[[0, 1, 2]])
    # print(C[0:3])
    # print(C.to_dense())
    # print(Ct.to_dense())

    # uvw = C.dot(xyz)

    # print(uvw)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ._core import *

from .linalg import *
from .matrices import *


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import angle_vectors_xy
from compas_tna.equilibrium.parallelisation import parallelise_edges


__all__ = [
    'horizontal_nodal',
]


def horizontal_nodal(form, force, alpha=100, kmax=100, callback=None):
    r"""Compute horizontal equilibrium using a node-per-node approach.

    Parameters
    ----------
    form : compas_tna.diagrams.FormDiagram
        The form diagram.
    force : compas_tna.diagrams.ForceDiagram
        The force diagram.
    alpha : float, optional
        Weighting factor for computation of the target vectors (the default is
        100.0, which implies that the target vectors are the edges of the form diagram).
        If 0.0, the target vectors are the edges of the force diagram.
    kmax : int, optional
       Maximum number of iterations.
       Default is ``100``.
    callback : callable, optional
        A callback function to be called at every iteration of the parallelisation algorithm.
        The callback should take the current iterand, the coordinates of the form diagram,
        and the coordinates of the force diagram as input parameters.
        Default is ``None``.

    Returns
    -------
    None

    Notes
    -----
    This function will update the form and force diagram instead of returning a result.
    The relationship between force densities (``q``), horizontal forces (``h``), and lengths (``l``)
    is the following:

    .. math::

        Q_{i} &= \frac{F_{i, thrust}}{L_{i, thrust}} \\
              &= \frac{H_{i, form}}{L_{i, form}} \\
              &= scale * \frac{L_{i, force}}{L_{i, form}}

    """
    alpha = float(alpha) / 100.0
    alpha = max(0, min(1, alpha))
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    k_i = form.key_index()
    i_nbrs = {k_i[key]: [k_i[nbr] for nbr in form.vertex_neighbors(key)] for key in form.vertices()}
    fixed = set(list(form.anchors()) + list(form.fixed()))
    fixed = [k_i[key] for key in fixed]
    xy = form.vertices_attributes('xy')

    edges = list(form.edges_where({'_is_edge': True}))
    uv_i = form.uv_index()
    ij_e = {(k_i[u], k_i[v]): index for (u, v), index in iter(uv_i.items())}
    lmin = form.edges_attribute('lmin', edges=edges)
    lmax = form.edges_attribute('lmax', edges=edges)
    hmin = form.edges_attribute('hmin', edges=edges)
    hmax = form.edges_attribute('hmax', edges=edges)
    edges = [[k_i[u], k_i[v]] for u, v in edges]
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _k_i = force.key_index()
    _i_nbrs = {_k_i[key]: [_k_i[nbr] for nbr in force.vertex_neighbors(key)] for key in force.vertices()}
    _fixed = list(force.fixed())
    _fixed = [_k_i[key] for key in _fixed]
    _xy = force.vertices_attributes('xy')

    _edges = force.ordered_edges(form)
    _uv_i = {uv: index for index, uv in enumerate(_edges)}
    _ij_e = {(_k_i[u], _k_i[v]): index for (u, v), index in iter(_uv_i.items())}
    _lmin = force.edges_attribute('lmin', edges=_edges)
    _lmax = force.edges_attribute('lmax', edges=_edges)
    _edges = [[_k_i[u], _k_i[v]] for u, v in _edges]
    scale = force.attributes.get('scale', 1.0)
    # --------------------------------------------------------------------------
    # rotate force diagram to make it parallel to the form diagram
    # use CCW direction (opposite of cycle direction)
    # --------------------------------------------------------------------------
    _x, _y = zip(*_xy)
    _xy[:] = [list(item) for item in zip([-_ for _ in _y], _x)]
    # --------------------------------------------------------------------------
    # make the diagrams parallel to a target vector
    # that is the (alpha) weighted average of the directions of corresponding
    # edges of the two diagrams
    # --------------------------------------------------------------------------
    uv = [[xy[j][0] - xy[i][0], xy[j][1] - xy[i][1]] for i, j in edges]
    _uv = [[_xy[j][0] - _xy[i][0], _xy[j][1] - _xy[i][1]] for i, j in _edges]
    lengths = [(dx**2 + dy**2)**0.5 for dx, dy in uv]
    forces = [(dx**2 + dy**2)**0.5 for dx, dy in _uv]
    # --------------------------------------------------------------------------
    # the target vectors
    # --------------------------------------------------------------------------
    form_targets = [[alpha * v[0] / l, alpha * v[1] / l] if l else [0, 0] for v, l in zip(uv, lengths)]
    force_targets = [[(1 - alpha) * v[0] / l, (1 - alpha) * v[1] / l] if l else [0, 0] for v, l in zip(_uv, forces)]
    targets = [[a[0] + b[0], a[1] + b[1]] for a, b in zip(form_targets, force_targets)]
    # --------------------------------------------------------------------------
    # proper force bounds
    # --------------------------------------------------------------------------
    hmin[:] = [_ / scale for _ in hmin]
    hmax[:] = [_ / scale for _ in hmax]
    _lmin[:] = [max(a, b) for a, b in zip(hmin, _lmin)]
    _lmax[:] = [min(a, b) for a, b in zip(hmax, _lmax)]
    # --------------------------------------------------------------------------
    # parallelise
    # --------------------------------------------------------------------------
    if alpha < 1:
        parallelise_edges(xy, edges, targets, i_nbrs, ij_e, fixed=fixed, kmax=kmax, lmin=lmin, lmax=lmax)
    if alpha > 0:
        parallelise_edges(_xy, _edges, targets, _i_nbrs, _ij_e, fixed=_fixed, kmax=kmax, lmin=_lmin, lmax=_lmax, callback=callback)
    # --------------------------------------------------------------------------
    # update the coordinate difference vectors
    # --------------------------------------------------------------------------
    uv = [[xy[j][0] - xy[i][0], xy[j][1] - xy[i][1]] for i, j in edges]
    _uv = [[_xy[j][0] - _xy[i][0], _xy[j][1] - _xy[i][1]] for i, j in _edges]
    lengths = [(dx**2 + dy**2)**0.5 for dx, dy in uv]
    forces = [(dx**2 + dy**2)**0.5 for dx, dy in _uv]
    # --------------------------------------------------------------------------
    # compute the force densities
    # --------------------------------------------------------------------------
    q = [f / l for f, l in zip(forces, lengths)]
    # --------------------------------------------------------------------------
    # rotate the force diagram 90 degrees in CW direction
    # this way the relation between the two diagrams is easier to read
    # --------------------------------------------------------------------------
    _x, _y = zip(*_xy)
    _xy[:] = [list(item) for item in zip(_y, [-_ for _ in _x])]
    # --------------------------------------------------------------------------
    # angle deviations
    # note that this does not account for flipped edges!
    # --------------------------------------------------------------------------
    angles = [angle_vectors_xy(uv[i], _uv[i], deg=True) for i in range(len(edges))]
    # --------------------------------------------------------------------------
    # update form
    # --------------------------------------------------------------------------
    for key in form.vertices():
        i = k_i[key]
        form.vertex_attributes(key, 'xy', xy[i])
    for uv in form.edges_where({'_is_edge': True}):
        i = uv_i[uv]
        form.edge_attributes(uv, ['q', '_f', '_l', '_a'], [q[i], forces[i], lengths[i], angles[i]])
    # --------------------------------------------------------------------------
    # update force
    # --------------------------------------------------------------------------
    for key in force.vertices():
        i = _k_i[key]
        force.vertex_attributes(key, 'xy', _xy[i])
    for (u, v) in force.edges():
        if (u, v) in _uv_i:
            i = _uv_i[(u, v)]
        else:
            i = _uv_i[(v, u)]
        force.edge_attributes((u, v), ['_l', '_a'], [forces[i], angles[i]])


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from numpy import array
from numpy import float64

from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

from compas.numerical import connectivity_matrix
from compas.numerical import normrow

from compas_tna.diagrams import FormDiagram

from compas_tna.utilities import LoadUpdater
from compas_tna.utilities import update_z


__all__ = [
    'vertical_from_zmax',
    'vertical_from_q',
    'vertical_from_zmax_proxy',
    'vertical_from_q_proxy'
]


def vertical_from_zmax_proxy(formdata, *args, **kwargs):
    form = FormDiagram.from_data(formdata)
    scale = vertical_from_zmax(form, *args, **kwargs)
    return form.to_data(), scale


def vertical_from_q_proxy(formdata, *args, **kwargs):
    form = FormDiagram.from_data(formdata)
    vertical_from_q(form, *args, **kwargs)
    return form.to_data()


def vertical_from_zmax(form, zmax, kmax=100, xtol=1e-2, rtol=1e-3, density=1.0, display=False):
    """For the given form and force diagram, compute the scale of the force
    diagram for which the highest point of the thrust network is equal to a
    specified value.

    Parameters
    ----------
    form : compas_tna.diagrams.formdiagram.FormDiagram
        The form diagram
    force : compas_tna.diagrams.forcediagram.ForceDiagram
        The corresponding force diagram.
    zmax : float
        The maximum height of the thrust network (the default is None, which
        implies that the maximum height will be equal to a quarter of the diagonal
        of the bounding box of the form diagram).
    kmax : int
        The maximum number of iterations for computing vertical equilibrium
        (the default is 100).
    tol : float
        The stopping criterion.
    density : float
        The density for computation of the self-weight of the thrust network
        (the default is 1.0). Set this to 0.0 to ignore self-weight and only
        consider specified point loads.
    display : bool
        If True, information about the current iteration will be displayed.
        Default is False.

    Returns
    -------
    float
        The scale of the forcedensities.

    """
    xtol2 = xtol ** 2
    # --------------------------------------------------------------------------
    # FormDiagram
    # --------------------------------------------------------------------------
    k_i = form.key_index()
    uv_i = form.uv_index()

    vcount = len(form.vertex)
    anchors = list(form.anchors())
    fixed = [k_i[key] for key in anchors]
    free = list(set(range(vcount)) - set(fixed))
    xyz = array(form.vertices_attributes('xyz'), dtype=float64)
    thick = array(form.vertices_attribute('t'), dtype=float64).reshape((-1, 1))
    p = array(form.vertices_attributes(('px', 'py', 'pz')), dtype=float64)

    edges = list(form.edges_where({'_is_edge': True}))
    q = array(form.edges_attribute('q', edges=edges), dtype=float64).reshape((-1, 1))
    edges = [(k_i[u], k_i[v]) for u, v in edges]

    C = connectivity_matrix(edges, 'csr')
    Ci = C[:, free]
    Cf = C[:, fixed]
    Cit = Ci.transpose()
    Ct = C.transpose()
    # --------------------------------------------------------------------------
    # original data
    # --------------------------------------------------------------------------
    p0 = array(p, copy=True)
    q0 = array(q, copy=True)
    # --------------------------------------------------------------------------
    # load updater
    # --------------------------------------------------------------------------
    update_loads = LoadUpdater(form, p0, thickness=thick, density=density)
    # --------------------------------------------------------------------------
    # scale to zmax
    # note that zmax should not exceed scale * diagonal
    # --------------------------------------------------------------------------
    scale = 1.0

    for k in range(kmax):
        if display:
            print(k)

        update_loads(p, xyz)

        q = scale * q0
        Q = diags([q.ravel()], [0])
        A = Cit.dot(Q).dot(Ci)
        b = p[free, 2] - Cit.dot(Q).dot(Cf).dot(xyz[fixed, 2])
        xyz[free, 2] = spsolve(A, b)
        z = max(xyz[free, 2])
        res2 = (z - zmax) ** 2

        if k > 10:
            if res2 < xtol2:
                break

        scale = scale * (z / zmax)
    # --------------------------------------------------------------------------
    # vertical
    # --------------------------------------------------------------------------
    q = scale * q0
    Q = diags([q.ravel()], [0])

    update_z(xyz, Q, C, p, free, fixed, update_loads, tol=rtol, kmax=kmax, display=display)
    # --------------------------------------------------------------------------
    # update
    # --------------------------------------------------------------------------
    f = q * normrow(C.dot(xyz))
    r = Ct.dot(Q).dot(C).dot(xyz) - p
    # --------------------------------------------------------------------------
    # form
    # --------------------------------------------------------------------------
    for vertex in form.vertices():
        index = k_i[vertex]
        form.vertex_attribute(vertex, 'z', xyz[index, 2])
        form.vertex_attributes(vertex, ['_rx', '_ry', '_rz'], r[index])

    for edge in form.edges_where({'_is_edge': True}):
        index = uv_i[edge]
        form.edge_attributes(edge, ['q', '_f'], [q[index, 0], f[index, 0]])

    return scale


def vertical_from_q(form, scale=1.0, density=1.0, kmax=100, tol=1e-3, display=False):
    r"""Compute vertical equilibrium from the force densities of the independent edges.

    Parameters
    ----------
    form : FormDiagram
        The form diagram
    scale : float
        The scale of the horizontal forces.
        Default is ``1.0``.
    density : float, optional
        The density for computation of the self-weight of the thrust network.
        Set this to 0.0 to ignore self-weight and only consider specified point loads.
        Default is ``1.0``.
    kmax : int, optional
        The maximum number of iterations for computing vertical equilibrium.
        Default is ``100``.
    tol : float
        The stopping criterion.
        Default is ``0.001``.
    display : bool
        Display information about the current iteration.
        Default is ``False``.

    Notes
    -----
    The force densities stored in the Form Diagram are the ratios of lengths of corresponding edges in the Form and Force Diagram.
    This means they are not yet scaled with the scale of the horizontal forces.
    The horizontal forces stored in the diagram are scaled.

    .. math::

        q_i &= scale * \frac{l_{i, force}}{l_{i, form}} \\
            &= \frac{h_{i, form}}{l_{i, form}} \\
            &= \frac{f_{i, thrust}}{l_{i, thrust}}

    """
    k_i = form.key_index()
    uv_i = form.uv_index()

    vcount = form.number_of_vertices()
    anchors = list(form.anchors())
    fixed = [k_i[key] for key in anchors]
    free = list(set(range(vcount)) - set(fixed))
    xyz = array(form.vertices_attributes('xyz'), dtype=float64)
    thick = array(form.vertices_attribute('t'), dtype=float64).reshape((-1, 1))
    p = array(form.vertices_attributes(('px', 'py', 'pz')), dtype=float64)

    edges = list(form.edges_where({'_is_edge': True}))
    q = array(form.edges_attribute('q', edges=edges), dtype=float64).reshape((-1, 1))
    edges = [(k_i[u], k_i[v]) for u, v in edges]

    C = connectivity_matrix(edges, 'csr')
    # --------------------------------------------------------------------------
    # original data
    # --------------------------------------------------------------------------
    p0 = array(p, copy=True)
    q0 = array(q, copy=True)
    # --------------------------------------------------------------------------
    # load updater
    # --------------------------------------------------------------------------
    update_loads = LoadUpdater(form, p0, thickness=thick, density=density)
    # --------------------------------------------------------------------------
    # update forcedensity based on given q[ind]
    # --------------------------------------------------------------------------
    q = scale * q0
    Q = diags([q.ravel()], [0])
    # --------------------------------------------------------------------------
    # compute vertical
    # --------------------------------------------------------------------------
    update_z(xyz, Q, C, p, free, fixed, update_loads, tol=tol, kmax=kmax, display=display)
    # --------------------------------------------------------------------------
    # update
    # --------------------------------------------------------------------------
    f = q * normrow(C.dot(xyz))
    r = C.transpose().dot(Q).dot(C).dot(xyz) - p
    # --------------------------------------------------------------------------
    # form
    # --------------------------------------------------------------------------
    for vertex in form.vertices():
        index = k_i[vertex]
        form.vertex_attribute(vertex, 'z', xyz[index, 2])
        form.vertex_attributes(vertex, ['_rx', '_ry', '_rz'], r[index])

    for edge in form.edges_where({'_is_edge': True}):
        index = uv_i[edge]
        form.edge_attribute(edge, '_f', f[index, 0])


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import angle_vectors_xy

from numpy import array
from numpy import float64
from numpy import where

from compas.numerical import connectivity_matrix
from compas.numerical import normrow
from compas.numerical import normalizerow

from compas_tna.diagrams import FormDiagram
from compas_tna.diagrams import ForceDiagram

from compas_tna.utilities import rot90
from compas_tna.utilities import apply_bounds
from compas_tna.utilities import parallelise_sparse
from compas_tna.utilities import parallelise_nodal


__all__ = [
    'horizontal_numpy',
    'horizontal_nodal_numpy',
    'horizontal_numpy_proxy',
    'horizontal_nodal_numpy_proxy'
]


def horizontal_numpy_proxy(formdata, forcedata, *args, **kwargs):
    form = FormDiagram.from_data(formdata)
    force = ForceDiagram.from_data(forcedata)
    horizontal_numpy(form, force, *args, **kwargs)
    return form.to_data(), force.to_data()


def horizontal_nodal_numpy_proxy(formdata, forcedata, *args, **kwargs):
    form = FormDiagram.from_data(formdata)
    force = ForceDiagram.from_data(forcedata)
    horizontal_nodal_numpy(form, force, *args, **kwargs)
    return form.to_data(), force.to_data()


def horizontal_numpy(form, force, alpha=100.0, kmax=100):
    r"""Compute horizontal equilibrium.

    Parameters
    ----------
    form : compas_tna.diagrams.formdiagram.FormDiagram
    force : compas_tna.diagrams.forcediagram.ForceDiagram
    alpha : float
        Weighting factor for computation of the target vectors (the default is
        100.0, which implies that the target vectors are the edges of the form diagram).
        If 0.0, the target vectors are the edges of the force diagram.
    kmax : int
       Maximum number of iterations (the default is 100).

    Notes
    -----
    This implementation is based on the following formulation

    .. math::

        \mathbf{C}^{T} \mathbf{C} \mathbf{xy} = \mathbf{C}^{T} \mathbf{t}

    with :math:`\mathbf{C}` the connectivity matrix and :math:`\mathbf{t}` the
    target vectors.

    """
    # --------------------------------------------------------------------------
    # alpha == 1 : form diagram fixed
    # alpha == 0 : force diagram fixed
    # --------------------------------------------------------------------------
    alpha = max(0., min(1., float(alpha) / 100.0))
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    k_i = form.key_index()
    uv_i = form.uv_index()
    fixed = set(list(form.anchors()) + list(form.fixed()))
    fixed = [k_i[key] for key in fixed]
    xy = array(form.vertices_attributes('xy'), dtype=float64)

    edges = list(form.edges_where({'_is_edge': True}))
    lmin = array(form.edges_attribute('lmin', edges=edges), dtype=float64).reshape((-1, 1))
    lmax = array(form.edges_attribute('lmax', edges=edges), dtype=float64).reshape((-1, 1))
    hmin = array(form.edges_attribute('hmin', edges=edges), dtype=float64).reshape((-1, 1))
    hmax = array(form.edges_attribute('hmax', edges=edges), dtype=float64).reshape((-1, 1))
    edges = [[k_i[u], k_i[v]] for u, v in edges]

    C = connectivity_matrix(edges, 'csr')
    Ct = C.transpose()
    CtC = Ct.dot(C)
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _k_i = force.key_index()
    _uv_i = force.uv_index(form=form)
    _fixed = list(force.fixed())
    _fixed = [_k_i[key] for key in _fixed]
    _fixed = _fixed or [0]
    _xy = array(force.vertices_attributes('xy'), dtype=float64)

    _edges = force.ordered_edges(form)
    _lmin = array(force.edges_attribute('lmin', edges=_edges), dtype=float64).reshape((-1, 1))
    _lmax = array(force.edges_attribute('lmax', edges=_edges), dtype=float64).reshape((-1, 1))
    _edges = [[_k_i[u], _k_i[v]] for u, v in _edges]

    _C = connectivity_matrix(_edges, 'csr')
    _Ct = _C.transpose()
    _Ct_C = _Ct.dot(_C)

    scale = force.attributes.get('scale', 1.0)
    # --------------------------------------------------------------------------
    # rotate force diagram to make it parallel to the form diagram
    # use CCW direction (opposite of cycle direction)
    # --------------------------------------------------------------------------
    _xy[:] = rot90(_xy, +1.0)
    # --------------------------------------------------------------------------
    # make the diagrams parallel to a target vector
    # that is the (alpha) weighted average of the directions of corresponding
    # edges of the two diagrams
    # --------------------------------------------------------------------------
    uv = C.dot(xy)
    _uv = _C.dot(_xy)
    l = normrow(uv)
    _l = normrow(_uv)
    t = alpha * normalizerow(uv) + (1 - alpha) * normalizerow(_uv)
    # proper bounds
    hmin /= scale
    hmax /= scale
    _lmin = where(hmin > _lmin, hmin, _lmin)
    _lmax = where(hmax < _lmax, hmax, _lmax)
    # parallelise
    # add the outer loop to the parallelise function
    for k in range(kmax):
        # apply length bounds
        apply_bounds(l, lmin, lmax)
        apply_bounds(_l, _lmin, _lmax)
        if alpha != 1.0:
            # if emphasis is not entirely on the form
            # update the form diagram
            xy = parallelise_sparse(CtC, Ct.dot(l * t), xy, fixed, 'CtC')
            uv = C.dot(xy)
            l = normrow(uv)
        if alpha != 0.0:
            # if emphasis is not entirely on the force
            # update the force diagram
            _xy = parallelise_sparse(_Ct_C, _Ct.dot(_l * t), _xy, _fixed, '_Ct_C')
            _uv = _C.dot(_xy)
            _l = normrow(_uv)
    # --------------------------------------------------------------------------
    # compute the force densities
    # --------------------------------------------------------------------------
    f = _l
    q = (f / l).astype(float64)
    # --------------------------------------------------------------------------
    # rotate the force diagram 90 degrees in CW direction
    # this way the relation between the two diagrams is easier to read
    # --------------------------------------------------------------------------
    _xy[:] = rot90(_xy, -1.0)
    # --------------------------------------------------------------------------
    # angle deviations
    # note that this does not account for flipped edges!
    # --------------------------------------------------------------------------
    a = [angle_vectors_xy(uv[i], _uv[i], deg=True) for i in range(len(edges))]
    # --------------------------------------------------------------------------
    # update form
    # --------------------------------------------------------------------------
    for key, attr in form.vertices(True):
        i = k_i[key]
        attr['x'] = xy[i, 0]
        attr['y'] = xy[i, 1]
    for (u, v), attr in form.edges_where({'_is_edge': True}, True):
        i = uv_i[(u, v)]
        attr['q'] = q[i, 0]
        attr['_f'] = f[i, 0]
        attr['_l'] = l[i, 0]
        attr['_a'] = a[i]
    # --------------------------------------------------------------------------
    # update force
    # --------------------------------------------------------------------------
    for key, attr in force.vertices(True):
        i = _k_i[key]
        attr['x'] = _xy[i, 0]
        attr['y'] = _xy[i, 1]
    for (u, v), attr in force.edges(True):
        if (u, v) in _uv_i:
            i = _uv_i[(u, v)]
        else:
            i = _uv_i[(v, u)]
        attr['_l'] = _l[i, 0]
        attr['_a'] = a[i]


def horizontal_nodal_numpy(form, force, alpha=100, kmax=100):
    """Compute horizontal equilibrium using a node-per-node approach.

    Parameters
    ----------
    form : compas_tna.diagrams.FormDiagram
    force : compas_tna.diagrams.ForceDiagram
    alpha : float
        Weighting factor for computation of the target vectors (the default is
        100.0, which implies that the target vectors are the edges of the form diagram).
        If 0.0, the target vectors are the edges of the force diagram.
    kmax : int
       Maximum number of iterations (the default is 100).

    """
    alpha = float(alpha) / 100.0
    alpha = max(0., min(1., alpha))
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    k_i = form.key_index()
    uv_i = form.uv_index()
    i_nbrs = {k_i[key]: [k_i[nbr] for nbr in form.vertex_neighbors(key)] for key in form.vertices()}
    ij_e = {(k_i[u], k_i[v]): index for (u, v), index in iter(uv_i.items())}
    fixed = set(list(form.anchors()) + list(form.fixed()))
    fixed = [k_i[key] for key in fixed]
    edges = [[k_i[u], k_i[v]] for u, v in form.edges_where({'_is_edge': True})]
    lmin = array([attr.get('lmin', 1e-7) for key, attr in form.edges_where({'_is_edge': True}, True)], dtype=float64).reshape((-1, 1))
    lmax = array([attr.get('lmax', 1e+7) for key, attr in form.edges_where({'_is_edge': True}, True)], dtype=float64).reshape((-1, 1))
    hmin = array([attr.get('hmin', 1e-7) for key, attr in form.edges_where({'_is_edge': True}, True)], dtype=float64).reshape((-1, 1))
    hmax = array([attr.get('hmax', 1e+7) for key, attr in form.edges_where({'_is_edge': True}, True)], dtype=float64).reshape((-1, 1))
    flipmask = array([1.0 if not attr['_is_tension'] else -1.0 for key, attr in form.edges_where({'_is_edge': True}, True)], dtype=float).reshape((-1, 1))
    xy = array(form.vertices_attributes('xy'), dtype=float64)
    C = connectivity_matrix(edges, 'csr')
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _k_i = force.key_index()
    _uv_i = force.uv_index(form=form)
    _i_nbrs = {_k_i[key]: [_k_i[nbr] for nbr in force.vertex_neighbors(key)] for key in force.vertices()}
    _ij_e = {(_k_i[u], _k_i[v]): index for (u, v), index in iter(_uv_i.items())}
    _fixed = list(force.fixed())
    _fixed = [_k_i[key] for key in _fixed]
    _fixed = _fixed or [0]

    _edges = force.ordered_edges(form)
    _xy = array(force.vertices_attributes('xy'), dtype=float64)
    _lmin = array([attr.get('lmin', 1e-7) for key, attr in force.edges(True)], dtype=float64).reshape((-1, 1))
    _lmax = array([attr.get('lmax', 1e+7) for key, attr in force.edges(True)], dtype=float64).reshape((-1, 1))
    _edges = [[_k_i[u], _k_i[v]] for u, v in _edges]

    _C = connectivity_matrix(_edges, 'csr')
    scale = force.attributes.get('scale', 1.0)
    # --------------------------------------------------------------------------
    # rotate force diagram to make it parallel to the form diagram
    # use CCW direction (opposite of cycle direction)
    # --------------------------------------------------------------------------
    _xy[:] = rot90(_xy, +1.0)
    # --------------------------------------------------------------------------
    # make the diagrams parallel to a target vector
    # that is the (alpha) weighted average of the directions of corresponding
    # edges of the two diagrams
    # --------------------------------------------------------------------------
    uv = flipmask * C.dot(xy)
    _uv = _C.dot(_xy)
    l = normrow(uv)
    _l = normrow(_uv)
    # --------------------------------------------------------------------------
    # the target vectors
    # --------------------------------------------------------------------------
    targets = alpha * normalizerow(uv) + (1 - alpha) * normalizerow(_uv)
    # --------------------------------------------------------------------------
    # proper force bounds
    # --------------------------------------------------------------------------
    hmin /= scale
    hmax /= scale
    _lmin = where(hmin > _lmin, hmin, _lmin)
    _lmax = where(hmax < _lmax, hmax, _lmax)
    # --------------------------------------------------------------------------
    # parallelise
    # --------------------------------------------------------------------------
    if alpha < 1:
        parallelise_nodal(xy, C, targets, i_nbrs, ij_e, fixed=fixed, kmax=kmax, lmin=lmin, lmax=lmax)
    if alpha > 0:
        parallelise_nodal(_xy, _C, targets, _i_nbrs, _ij_e, kmax=kmax, lmin=_lmin, lmax=_lmax)
    # --------------------------------------------------------------------------
    # update the coordinate difference vectors
    # --------------------------------------------------------------------------
    uv = C.dot(xy)
    _uv = _C.dot(_xy)
    l = normrow(uv)
    _l = normrow(_uv)
    # --------------------------------------------------------------------------
    # compute the force densities
    # --------------------------------------------------------------------------
    f = flipmask * _l
    q = (f / l).astype(float64)
    # --------------------------------------------------------------------------
    # rotate the force diagram 90 degrees in CW direction
    # this way the relation between the two diagrams is easier to read
    # --------------------------------------------------------------------------
    _xy[:] = rot90(_xy, -1.0)
    # --------------------------------------------------------------------------
    # angle deviations
    # note that this does not account for flipped edges!
    # --------------------------------------------------------------------------
    a = [angle_vectors_xy(uv[i], _uv[i], deg=True) for i in range(len(edges))]
    # --------------------------------------------------------------------------
    # update form
    # --------------------------------------------------------------------------
    for key, attr in form.vertices(True):
        i = k_i[key]
        attr['x'] = xy[i, 0]
        attr['y'] = xy[i, 1]
    for (u, v), attr in form.edges_where({'_is_edge': True}, True):
        i = uv_i[(u, v)]
        attr['q'] = q[i, 0]
        attr['_f'] = f[i, 0]
        attr['_l'] = l[i, 0]
        attr['_a'] = a[i]
    # --------------------------------------------------------------------------
    # update force
    # --------------------------------------------------------------------------
    for key, attr in force.vertices(True):
        i = _k_i[key]
        attr['x'] = _xy[i, 0]
        attr['y'] = _xy[i, 1]
    for (u, v), attr in force.edges(True):
        if (u, v) in _uv_i:
            i = _uv_i[(u, v)]
        else:
            i = _uv_i[(v, u)]
        attr['_l'] = _l[i, 0]
        attr['_a'] = a[i]


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from numpy import empty_like
from numpy.linalg import cond

from scipy.linalg import cho_factor
from scipy.linalg import cho_solve
from scipy.linalg import lstsq
from scipy.linalg import solve
from scipy.linalg import norm

from scipy.sparse.linalg import factorized

from compas.numerical import connectivity_matrix
from compas.numerical import normrow
from compas.numerical import chofactor
from compas.numerical import lufactorized
from compas.numerical import dof
from compas.numerical import rref
from compas.numerical import nonpivots
from compas.numerical import equilibrium_matrix


__all__ = [
    'parallelise',
    'parallelise_sparse',
    'parallelise_nodal',
    'rot90',
    'apply_bounds',
    'update_z',
    'update_q_from_qind',
    'form_count_dof',
    'form_identify_dof',
]


EPS = 1 / sys.float_info.epsilon


def parallelise(A, B, X, known, k=1, key=None):
    unknown = list(set(range(X.shape[0])) - set(known))
    A11 = A[unknown, :][:, unknown]
    A12 = A[unknown, :][:, known]
    b = B[unknown] - A12.dot(X[known])
    if cond(A11) < EPS:
        if key:
            Y = cho_solve(chofactor(A11, key), b)
        else:
            Y = cho_solve(cho_factor(A11), b)
        X[unknown] = Y
        return X
    Y = lstsq(A11, b)
    Y = Y[0]
    X[unknown] = Y
    return X


def parallelise_sparse(A, B, X, known, k=1, key=None):
    unknown = list(set(range(X.shape[0])) - set(known))
    A11 = A[unknown, :][:, unknown]
    A12 = A[unknown, :][:, known]
    b = B[unknown] - A12.dot(X[known])
    if key:
        solve = lufactorized(A11, key)
        Y = solve(b)
    else:
        solve = factorized(A11)
        Y = solve(b)
    X[unknown] = Y
    return X


def parallelise_nodal(xy, C, targets, i_nbrs, ij_e, fixed=None, kmax=100, lmin=None, lmax=None):
    fixed = fixed or []
    fixed = set(fixed)

    n = xy.shape[0]

    for k in range(kmax):
        xy0 = xy.copy()
        uv = C.dot(xy)
        l = normrow(uv)

        if lmin is not None and lmax is not None:
            apply_bounds(l, lmin, lmax)

        for j in range(n):
            if j in fixed:
                continue

            nbrs = i_nbrs[j]
            xy[j, :] = 0.0

            for i in nbrs:
                if (i, j) in ij_e:
                    e = ij_e[(i, j)]
                    t = targets[e]
                elif (j, i) in ij_e:
                    e = ij_e[(j, i)]
                    t = -targets[e]
                else:
                    continue

                xy[j] += xy0[i] + l[e, 0] * t

            # add damping factor?
            xy[j] /= len(nbrs)

        for (i, j) in ij_e:
            e = ij_e[(i, j)]

            if l[e, 0] == 0.0:
                a = xy[i]
                b = xy[j]
                c = 0.5 * (a + b)
                xy[i] = c[:]
                xy[j] = c[:]


def rot90(xy, zdir=1.0):
    temp = empty_like(xy)
    temp[:, 0] = - zdir * xy[:, 1]
    temp[:, 1] = + zdir * xy[:, 0]
    return temp


def apply_bounds(x, xmin, xmax):
    xsmall = x < xmin
    xbig = x > xmax
    x[xsmall] = xmin[xsmall]
    x[xbig] = xmax[xbig]


def update_z(xyz, Q, C, p, free, fixed, updateloads, tol=1e-3, kmax=100, display=False):
    Ci = C[:, free]
    Cf = C[:, fixed]
    Ct = C.transpose()
    Cit = Ci.transpose()
    A = Cit.dot(Q).dot(Ci)
    A_solve = factorized(A)
    B = Cit.dot(Q).dot(Cf)
    CtQC = Ct.dot(Q).dot(C)

    updateloads(p, xyz)

    for k in range(kmax):
        if display:
            print(k)

        xyz[free, 2] = A_solve(p[free, 2] - B.dot(xyz[fixed, 2]))

        updateloads(p, xyz)

        r = CtQC.dot(xyz[:, 2]) - p[:, 2]
        residual = norm(r[free])

        if residual < tol:
            break

    return residual


def update_q_from_qind(E, q, dep, ind):
    """Update the full set of force densities using the values of the independent edges.

    Parameters
    ----------
    E : sparse csr matrix
        The equilibrium matrix.
    q : array
        The force densities of the edges.
    dep : list
        The indices of the dependent edges.
    ind : list
        The indices of the independent edges.

    Returns
    -------
    None
        The force densities are modified in-place.

    Examples
    --------
    >>>

    """
    m = E.shape[0] - len(dep)
    qi = q[ind]
    Ei = E[:, ind]
    Ed = E[:, dep]
    if m > 0:
        Edt = Ed.transpose()
        A = Edt.dot(Ed).toarray()
        b = Edt.dot(Ei).dot(qi)
    else:
        A = Ed.toarray()
        b = Ei.dot(qi)
    if cond(A) > EPS:
        res = lstsq(-A, b)
        qd = res[0]
    else:
        qd = solve(-A, b)
    q[dep] = qd


# ==============================================================================
# Form Diagram Functions
# ==============================================================================


def form_count_dof(form):
    k2i = form.key_index()
    xyz = form.vertices_attributes('xyz')
    fixed = [k2i[key] for key in form.anchors()]
    free = list(set(range(form.number_of_vertices())) - set(fixed))
    edges = [(k2i[u], k2i[v]) for u, v in form.edges_where({'_is_edge': True})]
    C = connectivity_matrix(edges)
    E = equilibrium_matrix(C, xyz, free)
    return dof(E)


def form_identify_dof(form, **kwargs):
    algo = kwargs.get('algo') or 'sympy'
    k2i = form.key_index()
    xyz = form.vertices_attributes('xyz')
    fixed = [k2i[key] for key in form.anchors()]
    free = list(set(range(form.number_of_vertices())) - set(fixed))
    edges = [(k2i[u], k2i[v]) for u, v in form.edges_where({'_is_edge': True})]
    C = connectivity_matrix(edges)
    E = equilibrium_matrix(C, xyz, free)
    return nonpivots(rref(E, algo=algo))


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.datastructures import Mesh


__all__ = ['Diagram']


class Diagram(Mesh):
    """Base diagram implementing attributes shared between the form and force diagram."""


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.datastructures import network_find_cycles
from compas.datastructures import Network
from compas.utilities import pairwise

from compas_tna.diagrams import Diagram


__all__ = ['FormDiagram']


class FormDiagram(Diagram):
    r"""The ``FormDiagram`` defines a TNA form diagram.

    Notes
    -----
    A ``FormDiagram`` has the following constructor functions

    *   ``from_obj`` : Construct a diagram from the geometry described in an OBJ file. Only points, lines, and faces are taken into account.
    *   ``from_json`` : Construct a diagram from a JSON file containing a serialised "data" dictionary.
    *   ``from_lines`` : Construct a diagram from pairs of line start and end points.
    *   ``from_mesh`` : Construct a diagram from a Mesh.
    *   ``from_rhinomesh`` : Construct a diagram from a Rhino mesh.
    *   ``from_rhinosurface`` : Construct a diagram from a Rhino surface, using the U and V isolines.
    *   ``from_rhinolines`` : Construct a diagram from a selection of Rhino lines (i.e. curves of degree 1).

    Default vertex/edge/face attributes can be "public" or "protected".
    Protected attributes are usually only for internal use and should only be modified by the algorithms that rely on them.
    If you do change them, do so with care...
    Protected vertex attributes are: ``_sw, _rx, _ry, _rz``.
    Protected edge attributes are: ``_h, _f, _a, _is_edge, _is_tension``.
    Protected face attributes are: ``_is_loaded``.

    The ``FormDiagram`` is a mesh.
    Since edges are implicit objects in COMPAS meshes, those edges that are not relevant from a TNA perspective have to be marked as ``_is_edge=False``.
    Usually, the user should not have to worry about this.
    Furthermore, changing an edge to ``_is_edge=False`` requires an equivalent change in the force diagram.
    Therefore, the attribute ``_is_edge`` is marked as "protected".

    The ``FormDiagram`` holds the information of the form diagram as well as the corresponding thrust surface.
    This means that the form diagram contains both 2D and 3D geometry and force information.
    During calculations and manipulations related to horizontal equilibrium, only the 2D information is used.
    The relationship between force density, length, axial force, horizontal force of form diagram edges
    and length of the corresponding force edges is the following:

    .. math::

        Q_{i} &= \frac{F_{i}}{L_{i, thrust}} \\
              &= \frac{H_{i}}{L_{i, form}} \\
              &= scale * \frac{L_{i, force}}{L_{i, form}}

    """

    def __init__(self):
        super(FormDiagram, self).__init__()
        self.dual = None
        self.default_vertex_attributes.update({
            'x': 0.0,
            'y': 0.0,
            'z': 0.0,
            'px': 0.0,
            'py': 0.0,
            'pz': 0.0,
            't': 1.0,
            'is_anchor': False,
            'is_fixed': False,

            '_rx': 0.0,
            '_ry': 0.0,
            '_rz': 0.0,
        })
        self.default_edge_attributes.update({
            'q': 1.0,
            'lmin': 0.0,
            'lmax': 1e+7,
            'hmin': 0.0,
            'hmax': 1e+7,

            '_h': 0.0,
            '_f': 0.0,
            '_l': 0.0,
            '_a': 0.0,

            '_is_edge': True,
            '_is_tension': False
        })
        self.default_face_attributes.update({
            '_is_loaded': True
        })
        self.attributes.update({
            'name': 'FormDiagram',
        })

    def __str__(self):
        """Compile a mesh summary of the form diagram."""
        numv = self.number_of_vertices()
        nume = len(list(self.edges_where({'_is_edge': True})))
        numf = self.number_of_faces()
        vmin = self.vertex_min_degree()
        vmax = self.vertex_max_degree()
        fmin = self.face_min_degree()
        fmax = self.face_max_degree()
        return """
Form Diagram
============
name: {}
number of vertices: {}
number of (real) edges: {}
number of faces: {}
vertex degree: {}/{}
face degree: {}/{}
""".format(self.name, numv, nume, numf, vmin, vmax, fmin, fmax)

    @classmethod
    def from_lines(cls, lines, delete_boundary_face=True, precision=None, **kwargs):
        """Construct a FormDiagram from a list of lines described by start and end point coordinates.

        Parameters
        ----------
        lines : list
            A list of pairs of point coordinates.
        delete_boundary_face : bool, optional
            Set ``True`` to delete the face on the outside of the boundary, ``False`` to keep it.
            Default is ``True``.
        precision: str, optional
            The precision of the geometric map that is used to connect the lines.
            If not specified, the global precision stored in ``compas.PRECISION`` will be used.

        Returns
        -------
        FormDiagram
            A FormDiagram object.

        Examples
        --------
        >>> import compas
        >>> from compas.files import OBJ
        >>> from compas_tna.diagrams import FormDiagram
        >>> obj = OBJ(compas.get('lines.obj'))
        >>> vertices = obj.parser.vertices
        >>> edges = obj.parser.lines
        >>> lines = [(vertices[u], vertices[v]) for u, v in edges]
        >>> form = FormDiagram.from_lines(lines)
        """
        network = Network.from_lines(lines, precision=precision)
        points = network.to_points()
        cycles = network_find_cycles(network, breakpoints=network.leaves())
        form = cls.from_vertices_and_faces(points, cycles)
        if delete_boundary_face:
            form.delete_face(0)
        if 'name' in kwargs:
            form.name = kwargs['name']
        return form

    @classmethod
    def from_mesh(cls, mesh, **kwargs):
        """Construct a FormDiagram from a Mesh.

        Parameters
        ----------
        mesh : compas.datastructures.Mesh
            The mesh to be taken as reference.
            The keys of the faces and vertices of the base mesh will be kept.
            Only the XY coordinates per vertex are stored. Z = 0.0

        Returns
        -------
        FormDiagram
            A FormDiagram object.

        Examples
        --------
        >>> import compas
        >>> from compas.datastructures import Mesh
        >>> from compas_tna.diagrams import FormDiagram

        >>> mesh = Mesh.from_obj(compas.get('faces.obj'))
        >>> form = FormDiagram.from_mesh(mesh)
        >>> form.plot()
        """
        form = cls()
        for vkey, attr in mesh.vertices(True):
            form.add_vertex(key=vkey, x=attr['x'], y=attr['y'], z=0.0)
        for fkey in mesh.faces():
            form.add_face(vertices=mesh.face_vertices(fkey), fkey=fkey)
        if 'name' in kwargs:
            mesh.name = kwargs['name']
        return form

    @classmethod
    def from_rhinomesh(cls, guid, **kwargs):
        """Construct a FormDiagram from a Rhino mesh represented by a guid.

        Parameters
        ----------
        guid : str
            A globally unique identifier.

        Returns
        -------
        FormDiagram
            A FormDiagram object.

        Examples
        --------
        .. code-block:: python

            import compas_rhino
            from compas_tna.diagrams import FormDiagram

            guid = compas_rhino.select_mesh()
            form = FormDiagram.from_rhinomesh(guid)

        """
        from compas_rhino.geometry import RhinoMesh
        mesh = RhinoMesh.from_guid(guid).to_compas(cls)
        if 'name' in kwargs:
            mesh.name = kwargs['name']
        return mesh

    @classmethod
    def from_rhinosurface(cls, guid, **kwargs):
        """Construct a FormDiagram from a Rhino surface represented by its GUID.

        Parameters
        ----------
        guid : str
            A globally unique identifier.

        Returns
        -------
        FormDiagram
            A FormDiagram object.

        Examples
        --------
        .. code-block:: python

            import compas_rhino
            from compas_tna.diagrams import FormDiagram

            guid = compas_rhino.select_surface()
            form = FormDiagram.from_rhinosurface(guid)

        """
        from compas_rhino.geometry import RhinoSurface
        mesh = RhinoSurface.from_guid(guid).uv_to_compas(cls, **kwargs)
        if 'name' in kwargs:
            mesh.name = kwargs['name']
        return mesh

    @classmethod
    def from_rhinolines(cls, guids, delete_boundary_face=True, precision=None, **kwargs):
        """Construct a FormDiagram from a set of Rhino lines represented by their GUIDs.

        Parameters
        ----------
        guids : list
            A list of GUIDs.
        delete_boundary_face : bool, optional
            Set ``True`` to delete the face on the outside of the boundary, ``False`` to keep it.
            Default is ``True``.
        precision: str, optional
            The precision of the geometric map that is used to connect the lines.
            If not specified, the global precision stored in ``compas.PRECISION`` will be used.

        Returns
        -------
        FormDiagram
            A FormDiagram object.

        Examples
        --------
        .. code-block:: python

            import compas_rhino
            from compas_tna.diagrams import FormDiagram

            guids = compas_rhino.select_lines()
            form = FormDiagram.from_rhinolines(guids)

        """
        import compas_rhino
        lines = compas_rhino.get_line_coordinates(guids)
        mesh = FormDiagram.from_lines(lines, delete_boundary_face=delete_boundary_face, precision=precision, **kwargs)
        return mesh

    def uv_index(self):
        """Returns a dictionary that maps edge keys (i.e. pairs of vertex keys)
        to the corresponding edge index in a list or array of edges.

        Returns
        -------
        dict
            A dictionary of uv-index pairs.
        """
        return {(u, v): index for index, (u, v) in enumerate(self.edges_where({'_is_edge': True}))}

    def index_uv(self):
        """Returns a dictionary that maps edges in a list to the corresponding
        vertex key pairs.

        Returns
        -------
        dict
            A dictionary of index-uv pairs.
        """
        return dict(enumerate(self.edges_where({'_is_edge': True})))

    # --------------------------------------------------------------------------
    # dual and reciprocal
    # --------------------------------------------------------------------------

    def dual_diagram(self, cls):
        """Construct the dual of the FormDiagram.

        Parameters
        ----------
        cls : Mesh
            The type of the dual.

        Returns
        -------
        Mesh
            The dual as an instance of type ``cls``.

        Notes
        -----
        Construction of the dual diagram is based on the faces around the inner, free vertices of the form diagram.
        This means not only the vertices on the boundary are ignored, but also the vertices that are anchored.

        """
        dual = cls()
        fkey_centroid = {fkey: self.face_centroid(fkey) for fkey in self.faces()}
        inner = list(set(self.vertices()) - set(self.vertices_on_boundary()) - set(self.anchors()))
        vertices = {}
        faces = {}
        for key in inner:
            fkeys = self.vertex_faces(key, ordered=True)
            for fkey in fkeys:
                if fkey not in vertices:
                    vertices[fkey] = fkey_centroid[fkey]
            faces[key] = fkeys
        for key, (x, y, z) in vertices.items():
            dual.add_vertex(key, x=x, y=y, z=z)
        for fkey, vertices in faces.items():
            dual.add_face(vertices, fkey=fkey)
        return dual

    # --------------------------------------------------------------------------
    # vertices
    # --------------------------------------------------------------------------

    def leaves(self):
        """Vertices with degree 1.

        Returns
        -------
        generator
            A generator object for iteration over vertex keys that are leaves.
        """
        return self.vertices_where({'vertex_degree': 1})

    def corners(self):
        """Vertices with degree 2.

        Returns
        -------
        generator
            A generator object for iteration over vertex keys that are corners.
        """
        return self.vertices_where({'vertex_degree': 2})

    def anchors(self):
        """Vertices with ``is_anchor`` set to ``True``.

        Returns
        -------
        generator
            A generator object for iteration over vertex keys that are anchors.
        """
        return self.vertices_where({'is_anchor': True})

    def fixed(self):
        """Vertices with ``is_fixed`` set to ``True``.

        Returns
        -------
        generator
            A generator object for iteration over vertex keys that are fixed.
        """
        return self.vertices_where({'is_fixed': True})

    # def residual(self):
    #     R = 0
    #     for vertex in self.vertices_where({'is_anchor': False, 'is_fixed': False}):
    #         rx, ry, rz = self.vertex_attributes(vertex, ['_rx', '_ry', '_rz'])
    #         R += rx ** 2 + ry ** 2 + rz ** 2
    #     return sqrt(R)

    # --------------------------------------------------------------------------
    # helpers
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # postprocess
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # boundary conditions
    # --------------------------------------------------------------------------

    def update_boundaries(self):
        """Update the boundaries to add outside faces."""
        # mark all "anchored edges" as '_is_edge=False'
        # they will be ignored in any futher steps
        # this is what indierectly creates isolated vertices
        # and vor example the corner cutoffs in orthogonal grids
        for edge in self.edges():
            if all(self.vertices_attribute('is_anchor', keys=edge)):
                if self.is_edge_on_boundary(*edge):
                    self.edge_attribute(edge, '_is_edge', False)
        # delete isolated (corner) vertices
        # technically this can happen for vertices with degree of any kind
        for vertex in list(self.vertices_where({'vertex_degree': 2})):
            nbrs = self.vertex_neighbors(vertex)
            if all(not self.edge_attribute((vertex, nbr), '_is_edge') for nbr in nbrs):
                for nbr in nbrs:
                    face = self.halfedge[vertex][nbr]
                    if face is not None:
                        break
                vertices = self.face_vertices(face)
                after = nbr
                before = vertices[vertices.index(after) - 2]
                self.split_face(face, before, after)
                self.edge_attribute((before, after), '_is_edge', False)
                self.delete_vertex(vertex)
        # boundaries
        for boundary in self.vertices_on_boundaries():
            anchors = [vertex for vertex in boundary if self.vertex_attribute(vertex, 'is_anchor')]
            if len(anchors) == 0:
                # if the boundary contains no anchors
                # only an additional face has to be added
                # this tends to only be the case with openings/holes
                vertices = boundary
                self.add_face(vertices, _is_loaded=False)
            elif len(anchors) == 1:
                # if the boundary has exactly 1 anchor
                # the boundary just has to be cut at the anchor and pasted back together
                # and then a face has to be added
                i = boundary.index(anchors[0])
                vertices = boundary[i:] + boundary[:i]
                self.add_face(vertices, _is_loaded=False)
            else:
                # if the boundary has more than 1 anchor
                # split the boundary into segments between the anchors
                # and add a boundary face for every segment
                segments = []
                for start, end in pairwise(anchors + anchors[:1]):
                    i = boundary.index(start)
                    j = boundary.index(end)
                    if i < j:
                        segment = boundary[i:j + 1]
                    elif i > j:
                        segment = boundary[i:] + boundary[:j + 1]
                    else:
                        continue
                    segments.append(segment)
                # add outer faces
                for vertices in segments:
                    if len(vertices) < 3:
                        continue
                    self.add_face(vertices, _is_loaded=False)
                    self.edge_attribute((vertices[0], vertices[-1]), '_is_edge', False)

    # --------------------------------------------------------------------------
    # visualisation
    # --------------------------------------------------------------------------

    def plot(self):
        """Plot a form diagram with a plotter with all the default settings."""
        from compas_plotters import MeshPlotter
        plotter = MeshPlotter(self, figsize=(12, 8), tight=True)
        vertexcolor = {}
        vertexcolor.update({key: '#00ff00' for key in self.vertices_where({'is_fixed': True})})
        vertexcolor.update({key: '#ff0000' for key in self.vertices_where({'is_anchor': True})})
        plotter.draw_vertices(facecolor=vertexcolor)
        plotter.draw_edges(keys=list(self.edges_where({'_is_edge': True})))
        plotter.draw_faces(keys=list(self.faces_where({'_is_loaded': True})))
        plotter.show()

    def draw(self, layer=None, clear_layer=True, settings=None):
        """Draw the form diagram in Rhino.

        Parameters
        ----------
        layer : str, optional
            The layer in which the drawing should be contained.
        clear_layer : bool, optional
            Clear the layer if ``True``.
            Default is ``True``.
        settings : dict, optional
            A dictionary of settings overwriting the default settings of the artist.
        """
        from compas_tna.rhino import FormArtist
        artist = FormArtist(self, layer=layer, settings=settings)
        if clear_layer:
            artist.clear_layer()
        artist.draw()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    pass



"""
********************************************************************************
compas_tna.diagrams
********************************************************************************

.. currentmodule:: compas_tna.diagrams


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    FormDiagram
    ForceDiagram


Bases
=====

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Diagram

"""
from __future__ import absolute_import

from .diagram import *  # noqa: F401 F403
from .formdiagram import *  # noqa: F401 F403
from .forcediagram import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from copy import deepcopy

from compas.geometry import cross_vectors
from compas.geometry import length_vector
from compas.geometry import centroid_points
from compas.geometry import norm_vector

from compas_tna.equilibrium._alglib._core import xalglib


__all__ = ['vertical_from_q_alglib']


def vertical_from_q_alglib(form, scale=1.0, density=1.0, kmax=100, tol=1e-3):
    """"""
    key_index = form.key_index()
    xyz = form.vertices_attributes('xyz')
    loads = form.vertices_attributes(('px', 'py', 'pz'))
    n = form.number_of_vertices()
    fixed = list(set(list(form.anchors()) + list(form.fixed())))
    free = list(set(range(n)) - set(fixed))
    ni = len(free)
    nf = len(fixed)
    xyzf = [xyz[i] for i in fixed]
    selfweight = selfweight_calculator(form, density=density)
    adjacency = {}
    for key in form.vertices():
        nbrs = form.vertex_neighbors(key)
        adj = [key_index[nbr] for nbr in nbrs if form.edge_attribute((key, nbr), '_is_edge')]
        adjacency[key_index[key]] = adj
    ij_q = {uv: scale * form.edge_attribute(uv, 'q', 1.0) for uv in form.edges_where({'_is_edge': True})}
    ij_q.update({(v, u): q for (u, v), q in ij_q.items()})
    ij_q = {(key_index[u], key_index[v]): ij_q[u, v] for u, v in ij_q}
    nonzero_fixed, nonzero_free = nonzero(adjacency, fixed, free)
    CtQC = xalglib.sparsecreate(n, n)
    CitQCi = xalglib.sparsecreate(ni, ni)
    CitQCf = xalglib.sparsecreate(ni, nf)
    solver = xalglib.linlsqrcreate(ni, ni)
    update_matrices(adjacency, free, nonzero_fixed, nonzero_free, CtQC, CitQCf, CitQCi, ij_q)
    update_z(solver, xyz, xyzf, free, CtQC, CitQCf, CitQCi, selfweight=selfweight, kmax=kmax, tol=tol)
    p = deepcopy(loads)
    sw = selfweight(xyz)
    for i in range(len(p)):
        p[i][2] -= sw[i]
    rx, ry, rz = compute_residuals(xyz, p, n, CtQC)
    for key in form.vertices():
        index = key_index[key]
        form.vertex_attributes(key, 'xyz', xyz[index])
        form.vertex_attributes(key, 'rx', rx[index])
        form.vertex_attributes(key, 'ry', ry[index])
        form.vertex_attributes(key, 'rz', rz[index])
    for u, v in form.edges():
        l = form.edge_length(u, v)
        f = q * l
        form.edge_attributes((u, v), ('f', 'l'), (f, l))


# ==============================================================================
# helpers
# ==============================================================================


def selfweight_calculator(form, density=1.0):
    key_index = form.key_index()
    sw = [0] * form.number_of_vertices()
    rho = [attr['t'] * density for key, attr in form.vertices(True)]
    def calculate_selfweight(xyz):
        fkey_centroid = {fkey: form.face_centroid(fkey) for fkey in form.faces() if form.face_attribute(fkey, '_is_loaded')}
        for u in form.vertices():
            i = key_index[u]
            p0 = xyz[i]
            area = 0
            for v in form.halfedge[u]:
                j = key_index[v]
                p01 = subtract_vectors(xyz[j], p0)
                fkey = form.halfedge[u][v]
                if fkey in fkey_centroid:
                    p02 = subtract_vectors(fkey_centroid[fkey], p0)
                    area += length_vector(cross_vectors(p01, p02))
                fkey = form.halfedge[v][u]
                if fkey in fkey_centroid:
                    p03 = subtract_vectors(fkey_centroid[fkey], p0)
                    area += length_vector(cross_vectors(p01, p03))
            sw[i] = 0.25 * area * rho[i]
        return sw
    return calculate_selfweight


def nonzero(adjacency, fixed, free):
    n = len(adjacency)
    j_col_free = {value: index for index, value in enumerate(free)}
    j_col_fixed = {value: index for index, value in enumerate(fixed)}
    i_nonzero_free = {i: [] for i in range(n)}
    i_nonzero_fixed = {i: [] for i in range(n)}
    fixed = set(fixed)
    for i in range(n):
        if i in fixed:
            i_nonzero_fixed[i].append((i, j_col_fixed[i]))
        else:
            i_nonzero_free[i].append((i, j_col_free[i]))
        for j in adjacency[i]:
            if j in fixed:
                i_nonzero_fixed[i].append((j, j_col_fixed[j]))
            else:
                i_nonzero_free[i].append((j, j_col_free[j]))
    return i_nonzero_fixed, i_nonzero_free


def update_matrices(adjacency, free, nonzero_fixed, nonzero_free, CtQC, CitQCf, CitQCi, ij_q):
    xalglib.sparseconverttohash(CtQC)
    xalglib.sparseconverttohash(CitQCi)
    xalglib.sparseconverttohash(CitQCf)
    n = len(adjacency)
    ni = len(free)
    for i in range(n):
        Q = 0
        for j in adjacency[i]:
            q = ij_q[(i, j)]
            Q += q
            xalglib.sparseset(CtQC, i, j, -q)
        xalglib.sparseset(CtQC, i, i, Q)
    for row in range(ni):
        i = free[row]
        for j, col in nonzero_fixed[i]:
            xalglib.sparseset(CitQCf, row, col, xalglib.sparseget(CtQC, i, j))
        for j, col in nonzero_free[i]:
            xalglib.sparseset(CitQCi, row, col, xalglib.sparseget(CtQC, i, j))


def update_z(solver, xyz, xyzf, free, CtQC, CitQCf, CitQCi, selfweight, tol=1e-3, kmax=100):
    # solve A.x = b
    # with A = CitQCi
    #      b = pzi - CitQCf.zf
    #      x = zi
    xalglib.sparseconverttocrs(CitQCi)
    xalglib.sparseconverttocrs(CitQCf)
    xalglib.sparseconverttocrs(CtQC)
    n = len(xyz)
    ni = len(free)
    z = [z for _, _, z in xyz]
    zf = [z for _, _, z in xyzf]
    A = CitQCi
    b_ = xalglib.sparsemv(CitQCf, zf, [0] * ni)
    out = [0] * n
    for k in range(kmax):
        sw = selfweight(xyz)
        b = [- sw[i][2] - b_[i] for i in range(ni)]
        xalglib.linlsqrsolvesparse(solver, A, b)
        zi, _ = xalglib.linlsqrresults(solver)
        for i in range(ni):
            z[free[i]] = zi[i]
        rz = xalglib.sparsemv(CtQC, z, out)
        rz = [- sw[i][2] - rz[i] for i in range(n)]
        residual = norm([rz[free[i]] for i in range(ni)])
        if residual < tol:
            break
    for i in range(ni):
        xyz[free[i]][2] = zi[i]
    return residual


def compute_residuals(xyz, p, n, CtQC):
    # residual = CtQC.xyz - p
    xalglib.sparseconverttocrs(CtQC)
    x, y, z = zip(*xyz)
    x = list(x)
    y = list(y)
    z = list(z)
    out = [0] * n
    rx = xalglib.sparsemv(CtQC, x, out)
    rx = [p[i][0] - rx[i] for i in range(n)]
    ry = xalglib.sparsemv(CtQC, y, out)
    ry = [p[i][1] - ry[i] for i in range(n)]
    rz = xalglib.sparsemv(CtQC, z, out)
    rz = [p[i][2] - rz[i] for i in range(n)]
    return rx, ry, rz


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from numpy import array
from scipy.interpolate import griddata


__all__ = ['distribute_thickness']


def distribute_thickness(formdiagram):
    """Distribute thickness by interpolating provided values over the vertex grid."""
    points = []
    values = []
    xi = []
    index_key = {}
    index = 0
    for key in formdiagram.vertices():
        t = formdiagram.vertex_attribute(key, 't')
        xy = formdiagram.vertex_attributes(key, 'xy')
        if t:
            points.append(xy)
            values.append(t)
        else:
            xi.append(xy)
            index_key[index] = key
            index += 1
    points = array(points)
    values = array(values)
    xi = array(xi)
    t = griddata(points, values, xi, method='linear')
    t = t.flatten().tolist()
    for index, value in enumerate(t):
        key = index_key[index]
        formdiagram.vertex_attribute(key, 't', value)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_rhino.artists import MeshArtist


__all__ = ['ForceArtist']


class ForceArtist(MeshArtist):
    """Artist for force diagrams.

    Parameters
    ----------
    force : :class:`compas_tna.diagrams.ForceDiagram`
        A force diagram.
    layer : str, optional
        The layer in which the artist should draw.
    settings : dict, optional
        Visualisation settings.

    Attributes
    ----------
    settings : dict
        Visualisation settings.

    Examples
    --------
    .. code-block:: python

        from compas_tna.rhino import ForceArtist

        artist = ForceArtist(force, layer='TNA::ForceDiagram')
        artist.draw()

    """

    def __init__(self, force, layer=None, settings=None):
        super(ForceArtist, self).__init__(force, layer=layer)
        self.settings.update({
            'color.vertex': (255, 255, 255),
            'color.edge': (0, 0, 0),
            'color.face': (210, 210, 210),
            'show.vertices': True,
            'show.edges': True,
            'show.faces': False})
        if settings:
            self.settings.update(settings)

    @property
    def force(self):
        return self.datastructure

    def clear(self):
        super(ForceArtist, self).clear()

    def draw(self):
        """Draw the force diagram.

        Notes
        -----
        To change the way individual components are drawn, modify the settings dict of the artist.

        """
        self.clear()
        if self.layer:
            self.clear_layer()
        if self.settings['show.vertices']:
            vertexcolor = {key: self.settings['color.vertices'] for key in self.vertices()}
            self.draw_vertices(color=vertexcolor)
        if self.settings['show.edges']:
            self.draw_edges(
                color=self.settings['color.edges'])
        if self.settings['show.faces']:
            self.draw_faces(
                color=self.settings['color.faces'])
        self.redraw()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass



"""
********************************************************************************
compas_tna.utilities
********************************************************************************

.. currentmodule:: compas_tna.utilities


Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    parallelise
    parallelise_sparse
    parallelise_nodal
    rot90
    apply_bounds
    update_z
    update_q_from_qind
    distribute_thickness
    relax_boundary_openings


Proxies
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    relax_boundary_openings_proxy


"""
from __future__ import absolute_import

import compas

if not compas.IPY:
    from .loads import *  # noqa: F401 F403
    from .pattern import *  # noqa: F401 F403
    from .thickness import *  # noqa: F401 F403
    from .diagrams import *  # noqa: F401 F403


__all__ = [name for name in dir() if not name.startswith('_')]



"""
********************************************************************************
compas_tna.equilibrium
********************************************************************************

.. currentmodule:: compas_tna.equilibrium


Horizontal
==========

Algorithms related to horizontal equilibrium of the form diagram and thus of the
corresponding thrust network.

Horizontal equilibrium is established when the form and force diagram are reciprocal.
Two diagrams are said to be reciprocal if they are each other's dual and if all
corresponding edges are at any specific constant angle.

Note that for legibility, the form and force diagrams in ``compas_tna`` are drawn
such that corresponding edges of reciprocal diagrams are perpendicular.
However, the diagrams are made reciprocal by parallelising the corresponding edges.
The diagrams are thus rotated by 90 degrees and back at the beginning and end of
each of the horizontal equilibrium algorithms.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    horizontal_nodal
    horizontal_numpy
    horizontal_nodal_numpy


Proxies
-------

The ``_numpy`` variants of the above functions can be used directly in Rhino using an RPC server proxy.

Note that whenever the original algorithm expects a diagram instance as parameter,
the corresponding proxy function will expect the data representing that diagram
instead. The data representing the diagram is stored in the ``data`` attribute of the
diagram.

>>> tna = Proxy('compas_tna.equilibrium')
>>> horizontal = tna.horizontal_nodal_numpy_proxy

Once the proxy is created, it can be used like an actual function with the same signature.

>>> formdata, forcedata = horizontal(form.data, force.data)
>>> form.data = formdata
>>> force.data = forcedata

.. autosummary::
    :toctree: generated/
    :nosignatures:

    horizontal_numpy_proxy
    horizontal_nodal_numpy_proxy


Vertical
========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    vertical_from_zmax
    vertical_from_q

Proxies
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    vertical_from_zmax_proxy
    vertical_from_q_proxy


"""
from __future__ import absolute_import

import compas

from .horizontal import *  # noqa: F401 F403

if compas.IPY:
    pass
else:
    from .horizontal_numpy import *  # noqa: F401 F403
    from .vertical_numpy import *  # noqa: F401 F403


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from numpy import array
from numpy import float64

from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

from compas.numerical import connectivity_matrix
from compas.numerical import normrow

from compas_tna.utilities import LoadUpdater
from compas_tna.utilities import update_z


__all__ = [
    'scale_from_target',
]


EPS = 1 / sys.float_info.epsilon


def scale_from_target(form, zmax, kmax=100, xtol=1e-2, rtol=1e-3, density=1.0, display=False):
    """For the given form and force diagram, compute the scale of the force
    diagram for which the highest point of the thrust network is equal to a
    specified value.

    Parameters
    ----------
    form : compas_tna.diagrams.formdiagram.FormDiagram
        The form diagram
    force : compas_tna.diagrams.forcediagram.ForceDiagram
        The corresponding force diagram.
    zmax : float
        The maximum height of the thrust network (the default is None, which
        implies that the maximum height will be equal to a quarter of the diagonal
        of the bounding box of the form diagram).
    kmax : int
        The maximum number of iterations for computing vertical equilibrium
        (the default is 100).
    tol : float
        The stopping criterion.
    density : float
        The density for computation of the self-weight of the thrust network
        (the default is 1.0). Set this to 0.0 to ignore self-weight and only
        consider specified point loads.
    display : bool
        If True, information about the current iteration will be displayed.

    Returns
    -------
    float
        The scale of the forcedensities.

    """
    xtol2 = xtol ** 2
    # --------------------------------------------------------------------------
    # FormDiagram
    # --------------------------------------------------------------------------
    k_i = form.key_index()
    uv_i = form.uv_index()
    vcount = len(form.vertex)
    anchors = list(form.anchors())
    fixed = list(form.fixed())
    fixed = set(anchors + fixed)
    fixed = [k_i[key] for key in fixed]
    free = list(set(range(vcount)) - set(fixed))
    edges = [(k_i[u], k_i[v]) for u, v in form.edges_where({'is_edge': True})]
    xyz = array(form.vertices_attributes('xyz'), dtype=float64)
    thick = array(form.vertices_attribute('t'), dtype=float64).reshape((-1, 1))
    p = array(form.vertices_attributes(('px', 'py', 'pz')), dtype=float64)
    q = [attr.get('q', 1.0) for key, attr in form.edges_where({'is_edge': True}, True)]
    q = array(q, dtype=float64).reshape((-1, 1))
    C = connectivity_matrix(edges, 'csr')
    Ci = C[:, free]
    Cf = C[:, fixed]
    Cit = Ci.transpose()
    Ct = C.transpose()
    # --------------------------------------------------------------------------
    # original data
    # --------------------------------------------------------------------------
    p0 = array(p, copy=True)
    q0 = array(q, copy=True)
    # --------------------------------------------------------------------------
    # load updater
    # --------------------------------------------------------------------------
    update_loads = LoadUpdater(form, p0, thickness=thick, density=density)
    # --------------------------------------------------------------------------
    # scale to zmax
    # note that zmax should not exceed scale * diagonal
    # --------------------------------------------------------------------------
    scale = 1.0

    for k in range(kmax):
        if display:
            print(k)

        update_loads(p, xyz)

        q = scale * q0
        Q = diags([q.ravel()], [0])
        A = Cit.dot(Q).dot(Ci)
        b = p[free, 2] - Cit.dot(Q).dot(Cf).dot(xyz[fixed, 2])
        xyz[free, 2] = spsolve(A, b)
        z = max(xyz[free, 2])
        res2 = (z - zmax) ** 2

        if res2 < xtol2:
            break

        scale = scale * (z / zmax)
    # --------------------------------------------------------------------------
    # vertical
    # --------------------------------------------------------------------------
    q = scale * q0
    Q = diags([q.ravel()], [0])

    res = update_z(xyz, Q, C, p, free, fixed, update_loads, tol=rtol, kmax=kmax, display=display)
    # --------------------------------------------------------------------------
    # update
    # --------------------------------------------------------------------------
    l = normrow(C.dot(xyz))
    f = q * l
    r = Ct.dot(Q).dot(C).dot(xyz) - p
    # --------------------------------------------------------------------------
    # form
    # --------------------------------------------------------------------------
    for key, attr in form.vertices(True):
        index = k_i[key]
        attr['z'] = xyz[index, 2]
        attr['_rx'] = r[index, 0]
        attr['_ry'] = r[index, 1]
        attr['_rz'] = r[index, 2]
    for key, attr in form.edges_where({'is_edge': True}, True):
        index = uv_i[key]
        attr['_f'] = f[index, 0]

    return scale


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import compas

try:
    from compas.numerical._alglib._core import Array
    from compas.numerical._alglib._core import Zeros
    from compas.numerical._alglib._core import Diagonal
    from compas.numerical._alglib._core import SparseArray

    from compas.numerical._alglib._core.xalglib import sparsecreate
    from compas.numerical._alglib._core.xalglib import sparseset
    from compas.numerical._alglib._core.xalglib import sparseget
    from compas.numerical._alglib._core.xalglib import sparsegetnrows
    from compas.numerical._alglib._core.xalglib import sparsegetncols
    from compas.numerical._alglib._core.xalglib import sparseconverttocrs
    from compas.numerical._alglib._core.xalglib import sparseconverttohash
    from compas.numerical._alglib._core.xalglib import sparseishash

except ImportError:
    compas.raise_if_ironpython()


__all__ = [
    'connectivity_matrix',
    'sparse_connectivity_matrix',
]


def connectivity_matrix(edges, v):
    e = len(edges)
    C = Zeros((e, v))
    for i in range(e):
        C[i, edges[i][0]] = -1
        C[i, edges[i][1]] = +1
    return C


def sparse_connectivity_matrix(edges, v, rtype=None):
    """Construct a sparse connectivity matrix.

    Parameters:
        edges (list) : Pairs of vertex indices.
        n (int) : The number of vertices.
        rtype (str) : Optional. The return type. Default is `None`.

    >>> import compas
    >>> network = Network.from_obj(compas.get_data('lines.obj'))
    >>> key_index = dict((key, index) for index, key in network.vertices_enum())
    >>> edges = [(key_index[u], key_index[v]) for u, v in network.edges_iter()]
    >>> n = len(network)
    >>> C = connectivity_matrix(edges, n)

    """
    e = len(edges)
    shape = e, v
    rows, cols, data = [], [], []
    for i in range(e):
        rows.append(i)
        rows.append(i)
        cols.append(edges[i][0])
        cols.append(edges[i][1])
        data.append(-1)
        data.append(+1)
    C = SparseArray((rows, cols, data), shape)
    return C


# def laplacian_matrix(neighbors, rtype=None):
#     """Construct a sparse Laplacian matrix.

#     Parameters:
#         neighbors (list) : An adjacency list. For each vertex it should contain
#             a list of neighboring vertex indices.
#         rtype (str) : Optional. The return type. Default is `None`.

#     >>> import compas
#     >>> network = Network.from_obj(compas.get_data('lines.obj'))
#     >>> k_i = dict((key, index) for index, key in network.vertices_enum())
#     >>> neighbors = [[k_i[nbr] for nbr in network.neighbors(key)] for key in network]
#     >>> L = laplacian_matrix(neighbors)

#     """
#     n = len(neighbors)
#     L = sparsecreate(n, n)
#     for i in range(n):
#         nbrs = neighbors[i]
#         w = -1.0 / len(nbrs)
#         for j in nbrs:
#             sparseset(L, i, j, w)
#         sparseset(L, i, i, 1.0)
#     if rtype == 'crs':
#         sparseconverttocrs(L)
#     return L


# def overconstrained_laplacian_matrix(L, constraints, rtype=None):
#     if not sparseishash(L):
#         sparseconverttohash(L)
#     c = len(constraints)
#     m = sparsegetnrows(L)
#     n = sparsegetncols(L)
#     O = sparsecreate(m + c, n)
#     for i in range(m):
#         for j in range(n):
#             v = sparseget(L, i, j)
#             sparseset(O, i, j, v)
#     for i in range(c):
#         j = constraints[i]
#         sparseset(O, m + i, j, 1.0)
#     if rtype == 'crs':
#         sparseconverttocrs(O)
#     return O


# def edgeweighted_laplacian_matrix(neighbors, rtype=None):
#     """Construct an *edge-weighted* Laplacian matrix.

#     Parameters:
#         neighbors (list) : An adjacency list where every adjacent vertex is a
#             tuple of the vertex index and the weight of the corresponding edge.
#             An example of edge weights is force density.
#         rtype (str) : Optional. The return type. Default is `None`.

#     >>> import compas
#     >>> network = Network.from_obj(compas.get_data('lines.obj'))
#     >>> k_i = dict((key, index) for index, key in network.vertices_enum())
#     >>> uv_q = dict(((u, v), attr['q']) for u, v, attr in network.edges_iter(True))
#     >>> uv_q.update(((v, u), attr['q']) for u, v, attr in network.edges_iter(True))
#     >>> neighbors = [[(k_i[nbr], uv_q[(key, nbr)]) for nbr in network.neighbors(key)] for key in network]
#     >>> CtQC = edgeweighted_laplacian_matrix(neighbors)

#     """
#     n = len(neighbors)
#     CtQC = sparsecreate(n, n)
#     for i in range(n):
#         Q = 0
#         for j, q in neighbors[i]:
#             Q += q
#             sparseset(CtQC, i, j, -q)
#         sparseset(CtQC, i, i, Q)
#     if rtype == 'crs':
#         sparseconverttocrs(CtQC)
#     return CtQC


# ==============================================================================
# Debugging
# ==============================================================================

if __name__ == "__main__":

    import compas
    from compas.datastructures import Network

    network = Network.from_obj(compas.get_data('lines.obj'))

    key_index = network.key_index()
    edges = [(key_index[u], key_index[v]) for u, v in network.edges()]

    n = network.number_of_vertices()

    C = connectivity_matrix(edges, n)

    print(C)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

try:
    from compas.numerical._alglib._core import Array
    from compas.numerical._alglib._core import Zeros
    from compas.numerical._alglib._core import xalglib

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['solve', 'spsolve', 'splsqr']


class LinalgError(Exception):
    pass


# def multiply_matrix_matrix(A, B):
#     if not sparseiscrs(A):
#         sparseconverttocrs(A)
#     m = sparsegetnrows(A)
#     n = sparsegetncols(A)
#     o = 3
#     res = [[0] * o for i in range(m)]
#     res = sparsemm(A, B, n, res)
#     return res


# rmatrixsolvelu(m)(fast) => combine with rmatrixlu (does not use caching)
# rmatrixsolve(m)(fast)

def solve(A, b, fast=True):
    m, k = A.shape
    k, n = b.shape
    if m != k:
        raise LinalgError('Solver requires square matrix.')
    if not n:
        n = 1
    if n == 1:
        if fast:
            info, x = xalglib.rmatrixsolvefast(A.data, m, b.flatdata)
        else:
            info, rep, x = xalglib.rmatrixsolve(A.data, m, b.flatdata)
    else:
        if fast:
            x, info = xalglib.rmatrixsolvemfast(A.data, m, b.data, n)
        else:
            info, rep, x = xalglib.rmatrixsolvem(A.data, m, b.data, n, True)
    return Array(x, (m, n))


def spsolve(A, b):
    pass


def splsqr(A, b, x=None):
    m, n = A.shape
    n, k = b.shape
    if not k:
        k = 1
    problem = xalglib.linlsqrcreate(m, n)
    if k > 1:
        if not x:
            x = Zeros((m, k))
        M = A.matrix
        for dim in range(k):
            xalglib.linlsqrsolvesparse(problem, M, b[:, dim].data)
            result = xalglib.linlsqrresults(problem)
            x[:, dim] = result[0]
        return x
    if not x:
        x = Zeros((m, k))
    xalglib.linlsqrsolvesparse(problem, M, b.data)
    result = xalglib.linlsqrresults(problem)
    x[:, 0] = result[0]
    return x


def lu_solve():
    pass


def cho_solve():
    pass


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import midpoint_point_point_xy


__all__ = [
    'parallelise_edges',
]


def parallelise_edges(xy, edges, targets, i_nbrs, ij_e, fixed=None, kmax=100, lmin=None, lmax=None, callback=None):
    """Parallelise the edges of a mesh to given target vectors.

    Parameters
    ----------
    xy : list
        The XY coordinates of the vertices of the edges.
    edges : list
        The edges as pairs of indices in ``xy``.
    targets : list
        A target vector for every edge.
    i_nbrs : dict
        A list of neighbours per vertex.
    ij_e : dict
        An edge index per vertex pair.
    fixed : list, optional
        The fixed nodes of the mesh.
        Default is ``None``.
    kmax : int, optional
        Maximum number of iterations.
        Default is ``100``.
    lmin : list, optional
        Minimum length per edge.
        Default is ``None``.
    lmax : list, optional
        Maximum length per edge.
        Default is ``None``.
    callback : callable, optional
        A user-defined callback function to be executed after every iteration.
        Default is ``None``.

    Returns
    -------
    None

    Examples
    --------
    >>>
    """
    if callback:
        if not callable(callback):
            raise Exception('The provided callback is not callable.')

    fixed = fixed or []
    fixed = set(fixed)

    n = len(xy)

    for k in range(kmax):
        xy0 = [[x, y] for x, y in xy]
        uv = [[xy[j][0] - xy[i][0], xy[j][1] - xy[i][1]] for i, j in edges]
        lengths = [(dx**2 + dy**2)**0.5 for dx, dy in uv]

        if lmin:
            lengths[:] = [max(a, b) for a, b in zip(lengths, lmin)]

        if lmax:
            lengths[:] = [min(a, b) for a, b in zip(lengths, lmax)]

        for j in range(n):
            if j in fixed:
                continue

            nbrs = i_nbrs[j]
            x, y = 0.0, 0.0

            for i in nbrs:
                ax, ay = xy0[i]

                if (i, j) in ij_e:
                    e = ij_e[(i, j)]
                    l = lengths[e]  # noqa: E741
                    tx, ty = targets[e]
                    x += ax + l * tx
                    y += ay + l * ty

                else:
                    e = ij_e[(j, i)]
                    l = lengths[e]  # noqa: E741
                    tx, ty = targets[e]
                    x += ax - l * tx
                    y += ay - l * ty

            xy[j][0] = x / len(nbrs)
            xy[j][1] = y / len(nbrs)

        for (i, j) in ij_e:
            e = ij_e[(i, j)]

            if lengths[e] == 0.0:
                c = midpoint_point_point_xy(xy[i], xy[j])
                xy[i][:] = c[:][:2]
                xy[j][:] = c[:][:2]

        if callback:
            callback(k, xy, edges)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    import doctest
    doctest.testmod(globs=globals())



"""
********************************************************************************
compas_tna
********************************************************************************

.. currentmodule:: compas_tna


.. toctree::
    :maxdepth: 1

    compas_tna.diagrams
    compas_tna.equilibrium
    compas_tna.rhino
    compas_tna.utilities

"""
from __future__ import print_function

import os
import sys


__author__ = ['Tom Van Mele', ]
__copyright__ = 'Copyright 2017 - Block Research Group, ETH Zurich'
__license__ = 'MIT License'
__email__ = 'vanmelet@ethz.ch'

__version__ = '0.1.2'


PY3 = sys.version_info.major == 3

HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, '../../'))
DATA = os.path.abspath(os.path.join(HOME, 'data'))
DOCS = os.path.abspath(os.path.join(HOME, 'docs'))
TEMP = os.path.abspath(os.path.join(HOME, 'temp'))


def get(filename):
    """Get the full path to one of the sample data files.

    Parameters
    ----------
    filename : str
        The name of the data file.

    Returns
    -------
    str
        The full path to the specified file.

    Notes
    -----
    The file name should be specified relative to the ``compas_tna`` sample data folder.
    This folder is only locally available if you installed ``compas_tna`` from source,
    or if you are working directly with the source.

    In all other cases, the function will get the corresponding files direcly from
    the GitHub repo, at https://raw.githubusercontent.com/BlockResearchGroup/compas_tna/master/data.

    Examples
    --------
    The ``compas_tna.get`` function is meant to be used in combination with the static
    constructors of the data structures.

    >>> import compas_tna
    >>> from compas_tna.diagrams import FormDiagram
    >>> form = FormDiagram.from_obj(compas.get('faces.obj'))

    """
    filename = filename.strip('/')

    localpath = os.path.abspath(os.path.join(DATA, filename))

    if os.path.exists(localpath):
        return localpath
    else:
        return "https://raw.githubusercontent.com/BlockResearchGroup/compas_tna/master/data/{}".format(filename)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from numpy import zeros

from compas.geometry import length_vector
from compas.geometry import cross_vectors
from compas.numerical import face_matrix


__all__ = ['LoadUpdater']


class LoadUpdater(object):
    """Class for constructing a callable for updating loads when geometry and selfweight change.

    Parameters
    ----------
    mesh : :class:`compas_tna.diagrams.FormDiagram`
        A form diagram mesh.
    p0 : ndarray (number_of_vertices x 3)
        The additional (fixed) loads at the vertices.
    thickness : ndarray (number_of_vertices x 1) or float, optional
        A thickness value per vertex.
        Or a single constant thickness value.
        Default is a constant thickness of ``1.0``.
    density : float, optional
        The density for selfweight calculations.
        Default is ``1.0``.

    Examples
    --------
    >>> import numpy
    >>> xyz = numpy.array(form.vertices_attributes('xyz'))
    >>> p = numpy.array(form.vertices_attributes(['px', 'py', 'pz']))
    >>> p0 = p.copy()

    Construct a load updater with the (additional) point loads applied at the vertices.
    Note that these additional loads are often zero.

    >>> updateloads = LoadUpdater(form, p0)

    After the geometry changes, update the loads by recomputing selfweight.

    >>> updateloads(p, xyz)
    """

    def __init__(self, mesh, p0, thickness=1.0, density=1.0, live=0.0):
        self.mesh = mesh
        self.p0 = p0
        self.thickness = thickness
        self.density = density
        self.live = live
        self.key_index = mesh.key_index()
        self.fkey_index = {fkey: index for index, fkey in enumerate(mesh.faces())}
        self.is_loaded = {fkey: mesh.face_attribute(fkey, '_is_loaded') for fkey in mesh.faces()}
        self.F = self._face_matrix()

    def __call__(self, p, xyz):
        ta = self._tributary_areas(xyz)
        sw = ta * self.thickness * self.density + ta * self.live
        p[:, 2] = self.p0[:, 2] + sw[:, 0]

    def _face_matrix(self):
        face_vertices = [None] * self.mesh.number_of_faces()
        for fkey in self.mesh.faces():
            face_vertices[self.fkey_index[fkey]] = [self.key_index[key] for key in self.mesh.face_vertices(fkey)]
        return face_matrix(face_vertices, rtype='csr', normalize=True)

    def _tributary_areas(self, xyz):
        mesh = self.mesh
        key_index = self.key_index
        fkey_index = self.fkey_index
        is_loaded = self.is_loaded
        C = self.F.dot(xyz)
        areas = zeros((xyz.shape[0], 1))
        for u in mesh.vertices():
            p0 = xyz[key_index[u]]
            a = 0
            for v in mesh.halfedge[u]:
                p1 = xyz[key_index[v]]
                p01 = p1 - p0
                fkey = mesh.halfedge[u][v]
                if fkey is not None and is_loaded[fkey]:
                    p2 = C[fkey_index[fkey]]
                    a += 0.25 * length_vector(cross_vectors(p01, p2 - p0))
                fkey = mesh.halfedge[v][u]
                if fkey is not None and is_loaded[fkey]:
                    p3 = C[fkey_index[fkey]]
                    a += 0.25 * length_vector(cross_vectors(p01, p3 - p0))
            areas[key_index[u]] = a
        return areas


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino

from compas.geometry import scale_vector
from compas.geometry import length_vector
from compas.geometry import add_vectors
from compas.utilities import i_to_green

from compas_rhino.artists import MeshArtist


__all__ = ['FormArtist']


class FormArtist(MeshArtist):
    """Artist for form diagrams.

    Parameters
    ----------
    form : :class:`compas_tna.diagrams.FormDiagram`
        A form diagram.
    layer : str, optional
        The layer in which the artist should draw.
    settings : dict, optional
        Visualisation settings.

    Attributes
    ----------
    settings : dict
        Visualisation settings.

    Examples
    --------
    .. code-block:: python

        from compas_tna.rhino import FormArtist

        artist = FormArtist(form, layer='TNA::FormDiagram')
        artist.draw()

    """

    def __init__(self, form, layer=None, settings=None):
        super(FormArtist, self).__init__(form, layer=layer)
        self.settings.update({
            'color.vertex': (255, 255, 255),
            'color.vertex:is_fixed': (0, 0, 255),
            'color.vertex:is_anchor': (255, 0, 0),
            'color.edge': (0, 0, 0),
            'color.face': (210, 210, 210),
            'color.reaction': (0, 255, 0),
            'color.residual': (0, 255, 255),
            'color.load': (0, 255, 0),
            'color.selfweight': (0, 0, 255),
            'color.force': (0, 0, 255),
            'scale.reaction': 1.0,
            'scale.residual': 1.0,
            'scale.load': 1.0,
            'scale.force': 1.0,
            'scale.selfweight': 1.0,
            'tol.reaction': 1e-3,
            'tol.residual': 1e-3,
            'tol.load': 1e-3,
            'tol.force': 1e-3,
            'tol.selfweight': 1e-3,
            'show.vertices': True,
            'show.edges': True,
            'show.faces': True,
            'show.loads': False,
            'show.reactions': False,
            'show.residuals': False,
            'show.forces': False,
            'show.angles': False})
        if settings:
            self.settings.update(settings)

    @property
    def form(self):
        return self.mesh

    def clear(self):
        super(FormArtist, self).clear()

    def draw(self):
        """Draw the form diagram.

        Notes
        -----
        To change the way individual components are drawn, modify the settings dict of the artist.

        """
        self.clear()
        if self.layer:
            self.clear_layer()
        if self.settings['show.vertices']:
            vertexcolor = {key: self.settings['color.vertices'] for key in self.vertices()}
            vertexcolor.update({key: self.settings['color.vertices:is_fixed'] for key in self.vertices_where({'is_fixed': True})})
            vertexcolor.update({key: self.settings['color.vertices:is_anchor'] for key in self.vertices_where({'is_anchor': True})})
            self.draw_vertices(color=vertexcolor)
        if self.settings['show.edges']:
            self.draw_edges(
                keys=list(self.edges_where({'_is_edge': True})),
                color=self.settings['color.edges'])
        if self.settings['show.faces']:
            self.draw_faces(
                keys=list(self.faces_where({'_is_loaded': True})),
                color=self.settings['color.faces'])
        if self.settings['show.forces']:
            self.draw_forces()
        if self.settings['show.reactions']:
            self.draw_reactions()
        if self.settings['show.angles']:
            self.draw_angles()
        self.redraw()

    def draw_loads(self, scale=None, color=None):
        """Draw the loads.

        Parameters
        ----------
        scale : float, optional
            Scaling factor for the load vectors.
            Default is the value from the settings.
        color : tuple, optional
            RGB color components for load vectors.
            Default is the value from the settings.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        lines = []
        color = color or self.settings['color.load']
        scale = scale or self.settings['scale.load']
        tol = self.settings['tol.load']
        for key in self.form.vertices_where({'is_anchor': False}):
            load = self.form.vertex_attributes(key, ['px', 'py', 'pz'])
            load = scale_vector(load, scale)
            if length_vector(load) < tol:
                continue
            sp = self.form.vertex_coordinates(key)
            ep = add_vectors(sp, load)
            lines.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'end',
                'name': "{}.load.{}".format(self.form.name, key)})
        guids = compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids

    def draw_selfweight(self, scale=None, color=None):
        """Draw the selfweight.

        Parameters
        ----------
        scale : float, optional
            Scaling factor for the selfweight vectors.
            Default is the value from the settings.
        color : tuple, optional
            RGB color components for selfweight vectors.
            Default is the value from the settings.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        lines = []
        color = color or self.settings['color.selfweight']
        scale = scale or self.settings['scale.selfweight']
        tol = self.settings['tol.selfweight']
        tol2 = tol ** 2
        for key in self.form.vertices_where({'is_anchor': False}):
            t = self.form.vertex_attribute(key, 't')
            a = self.form.vertex_area(key)
            sp = self.form.vertex_coordinates(key)
            dz = scale * t * a
            if dz ** 2 < tol2:
                continue
            ep = [sp[0], sp[1], sp[2] - dz]
            lines.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'end',
                'name': "{}.selfweight.{}".format(self.form.name, key)})
        guids = compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids

    def draw_reactions(self, scale=None, color=None):
        """Draw the reactions.

        Parameters
        ----------
        scale : float, optional
            Scaling factor for the reaction vectors.
            Default is the value from the settings.
        color : tuple, optional
            RGB color components for reaction vectors.
            Default is the value from the settings.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        lines = []
        color = color or self.settings['color.reaction']
        scale = scale or self.settings['scale.reaction']
        tol = self.settings['tol.reaction']
        for key in self.form.vertices_where({'is_anchor': True}):
            reaction = self.form.vertex_attributes(key, ['_rx', '_ry', '_rz'])
            reaction = scale_vector(reaction)
            if length_vector(reaction) < tol:
                continue
            sp = self.form.vertex_coordinates(key)
            ep = add_vectors(sp, reaction)
            lines.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'start',
                'name': "{}.reaction.{}".format(self.form.name, key)})
        guids = compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids

    def draw_forces(self, scale=None, color=None):
        """Draw the forces.

        Parameters
        ----------
        scale : float, optional
            Scaling factor for the force pipes.
            Default is the value from the settings.
        color : tuple, optional
            RGB color components for force pipes.
            Default is the value from the settings.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        lines = []
        color = color or self.settings['color.force']
        scale = scale or self.settings['scale.force']
        tol = self.settings['tol.force']
        for u, v in self.form.edges_where({'_is_edge': True}):
            force = self.form.edge_attribute((u, v), '_f')
            sp, ep = self.form.edge_coordinates(u, v)
            radius = scale * force
            if radius < tol:
                continue
            lines.append({
                'start': sp,
                'end': ep,
                'radius': radius,
                'color': color,
                'name': "{}.force.{}-{}".format(self.form.name, u, v)})
        guids = compas_rhino.draw_cylinders(lines, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids

    def draw_residuals(self, scale=None, color=None):
        """Draw the residual forces.

        Parameters
        ----------
        scale : float, optional
            Scaling factor for the force vectors.
            Default is the value from the settings.
        color : tuple, optional
            RGB color components for force vectors.
            Default is the value from the settings.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        lines = []
        color = color or self.settings['color.residual']
        scale = scale or self.settings['scale.residual']
        tol = self.settings['tol.residual']
        for key in self.form.vertices_where({'is_anchor': False}):
            residual = self.form.vertex_attributes(key, ['_rx', '_ry', '_rz'])
            residual = scale_vector(residual, scale)
            if length_vector(residual) < tol:
                continue
            sp = self.form.vertex_coordinates(key)
            ep = add_vectors(sp, residual)
            lines.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'start',
                'name': "{}.residual.{}".format(self.form.name, key)})
        guids = compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids

    def draw_angles(self, tol=5.0):
        """Draw the angle deviations.

        Parameters
        ----------
        tol : float, optional
            Tolerance value for angle deviations.
            Default value is ``5.0``.

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        labels = []
        for u, v in self.form.edges_where({'_is_edge': True}):
            a_rad = self.form.edge_attribute((u, v), '_a')
            a_deg = 180 * a_rad / 3.14159
            if a_deg > tol:
                color = i_to_green(a_rad / tol)
                labels.append({
                    'pos': self.form.edge_midpoint(u, v),
                    'text': "{:.1f}".format(a_deg),
                    'color': color,
                    'name': "{}.angle.{}-{}".format(self.form.name, u, v)})
        guids = compas_rhino.draw_labels(labels, layer=self.layer, clear=False, redraw=False)
        self.guids += guids
        return guids


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass
